"""View parent object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.application.dialogs.ChildDialogGenericView import ChildDialogGenericView
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
from pollenisatorgui.core.forms.formpanel import FormPanel
import pollenisatorgui.core.components.settings as settings
from pollenisatorgui.core.application.dialogs.ChildDialogToast import ChildDialogToast
import tkinter.messagebox
from tkinter import ttk
from tkinter import TclError
from customtkinter import *
from PIL import Image
import tkinter as tk
import pollenisatorgui.core.components.utils as utils


class ViewElement(object):
    """
    Defines a basic view to be inherited. Those functions are generic entry points to models.
    Most of them should not be redefined in other Views.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory
        cachedClassIcon: a cached loaded PIL image icon of ViewElement.icon. Starts as None.
    """
    icon = 'undefined.png'
    cachedClassIcon = None
    loadingicon = 'loading.png'
    cachedloadingicon = None
    multiview_class = None

    def __init__(self, appTw, appViewFrame, mainApp, controller):
        """Constructor
        Args:
            appTw: a PollenisatorTreeview instance to put this view in
            appViewFrame: an view frame to build the forms in.
            mainApp: the Application instance
            controller: a CommandController for this view.
        """
        self.appliTw = appTw
        self.appliViewFrame = appViewFrame
        self.mainApp = mainApp
        self.controller = controller
        self._form = None
        self._opened = False

    @property
    def form(self):
        """
        Returns the form of this view. If it is not yet created, it will create it.
        """
        if self._form is None:
            self._form = FormPanel()
        return self._form

    @classmethod
    def getClassIcon(cls):
        """
        Load the class icon in cache if it is not yet done, and returns it

        Return:
            Returns the ImageTk.PhotoImage icon representing this class .
        """
        from PIL import Image, ImageTk
        if cls.cachedClassIcon is None:
            path = utils.getIcon(cls.icon)
            cls.cachedClassIcon = ImageTk.PhotoImage(Image.open(path))
        return cls.cachedClassIcon
    
    @classmethod
    def getLoadingIcon(cls):
        """
        Load the class icon in cache if it is not yet done, and returns it

        Return:
            Returns the ImageTk.PhotoImage icon representing this class .
        """
        from PIL import Image, ImageTk
        if cls.cachedloadingicon is None:
            path = utils.getIcon(cls.loadingicon)
            cls.cachedloadingicon = ImageTk.PhotoImage(Image.open(path))
        return cls.cachedloadingicon

    def getIcon(self):
        """
        Load the object icon in cache if it is not yet done, and returns it

        Return:
            Returns the icon representing this object.
        """
        return self.__class__.getClassIcon()

    def addChildrenBaseNodes(self, newNode):
        """
        Add to the given node from a treeview the mandatory childrens.
        Will be redefined in children.

        Args:
            newNode: the newly created node we want to add children to.
        """
        # pass

    def opened(self):
        """Callback called when the view is opened"""
        if self._opened:
            return
        self._opened = True
        try:
            icon = self.appliTw.item(str(self.controller.getDbId()))["image"]
            self.appliTw.item(str(self.controller.getDbId()), image=ViewElement.getLoadingIcon())
            self.appliTw.update()
            self.appliTw.delete(str(self.controller.getDbId())+"|<Empty>")
        except tk.TclError as e:
            pass
        self._insertChildren()
        for module in self.mainApp.modules:
            if hasattr(module["object"], "_insertChildren"):
                module["object"]._insertChildren(self.controller.model.coll_name, self.controller.getData())
        try:
            self.appliTw.item(str(self.controller.getDbId()), image=icon)
        except tk.TclError:
            pass
        self.appliTw.update()
        
    def delete(self, _event=None, showWarning=True):
        """
        Entry point to the model doDelete function.

        Args:
            _event: automatically filled if called by an event. Not used
            showWarning: a boolean. If true, the user will be asked a confirmation before supression. Default to True.
        """
        ret = True
        if showWarning:
            dialog = ChildDialogQuestion(self.mainApp,
                                     "DELETE WARNING", "Becareful for you are about to delete this entry and there is no turning back.", ["Delete", "Cancel"])
            self.mainApp.wait_window(dialog.app)
            if dialog.rvalue != "Delete":
                return
        self.controller.doDelete()

    def getAdditionalContextualCommands(self):
        return {}

    def update(self, event=None):
        """
        Entry point to the model doUpdate function.

        Args:
            event: automatically filled if called by an event. Holds info on update clicked widget.
        Returns:
            * a boolean to shwo success or failure
            * an empty message on success, an error message on failure
        """
        res, msg = self.form.checkForm()
        if(res):
            form_values = self.form.getValue()
            form_values_as_dicts = ViewElement.list_tuple_to_dict(form_values)
            self.controller.doUpdate(form_values_as_dicts)
            try:
                caller = self.update_btn.btn
                toast = ChildDialogToast(self.appliViewFrame, "Done" , x=caller.winfo_rootx(), y=caller.winfo_rooty()+caller.winfo_reqheight(), width=caller.winfo_reqwidth())
                toast.show()
            except Exception:
                pass
            return True, ""
        else:
            tkinter.messagebox.showwarning(
                "Form not validated", msg, parent=self.appliViewFrame)
            return False, msg
        
    def _insertChildren(self):
        return

    def insert(self, _event=None):
        """
        Entry point to the model doInsert function.

        Args:
            _event: automatically filled if called by an event. Not used
        Returns:
            * a boolean to shwo success or failure
            * an empty message on success, an error message on failure
        """
        res, msg = self.form.checkForm()
        if(res):
            form_values = self.form.getValue()
            form_values_as_dicts = ViewElement.list_tuple_to_dict(form_values)
            res, msg = self.controller.doInsert(form_values_as_dicts)
            if not res:
                msg = "This element cannot be inserted, check for conflicts with existings elements."
                tkinter.messagebox.showerror(
                    "Insertion failed", msg, parent=self.appliViewFrame)
                return False, msg
            else:
                if isinstance(msg, tuple):
                    nbErrors = msg[0]
                    iid = msg[1]
                else:
                    iid = None
                    nbErrors = msg
                if nbErrors > 0:
                    msg = str(len(
                        res))+" were inserted, "+str(nbErrors)+" were not to avoid conflicts or out of wave elements."
                    tkinter.messagebox.showwarning(
                        "Insertion succeeded with warnings", msg, parent=self.appliViewFrame)
                    return True, msg
                else:
                    return True, str(iid)
        else:
            tkinter.messagebox.showwarning(
                "Form not validated", msg, parent=self.appliViewFrame)
            return False, msg

    def tagClicked(self, name):
        """Callback intermediate for tag clicked
        Ensure that the tag name clicked is added to View item
        Args:
            name: a tag name
        """
        return lambda _event: self.tagButtonClicked(name)
    
    def tagButtonClicked(self, name):
        """Callback for tag button clicked
        Ensure that the tag name clicked is set to View item
        Args:
            name: a tag name
        """
        self.controller.setTags([name])

    def completeModifyWindow(self, editable=True, addTags=True):
        """
        Add the buttons for an update window.
            -Submit button that validates the form with the update function.
            -Delete button that asks the user to delete the object with the delete function.
        """
        pan = self.form.addFormPanel()
        self.delete_image = CTkImage(Image.open(utils.getIcon("delete.png")))
        self.save_image = CTkImage(Image.open(utils.getIcon("save.png")))
        if editable:
            self.update_btn = pan.addFormButton("Submit", self.update, image=self.save_image)
            pan.addFormButton("Delete", self.delete, image=self.delete_image,
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
            # if addTags:
            #     registeredTags = settings.Settings.getTags()
            #     keys = list(registeredTags.keys())
            #     column = 0
            #     item_no = 0
            #     listOfLambdas = [self.tagClicked(keys[i]) for i in range(len(keys))]
            #     for registeredTag, color in registeredTags.items():
            #         if column == 0:
            #             panTags = self.form.addFormPanel(pady=0)
            #         s = ttk.Style(self.mainApp)
            #         try: # CHECK IF COLOR IS VALID
            #             if color == "transparent":
            #                 color = "white"
            #             CTkLabel(self.mainApp, fg_color=color)
            #         except tkinter.TclError as e:
            #             #color incorrect
            #             color = "gray97"
            #         btn_tag = panTags.addFormButton(registeredTag, listOfLambdas[item_no],  side="left", padx=1, pady=0)
            #         btn_tag.configure(fg_color=color, border_width=1, text_color="black")
            #         column += 1
            #         item_no += 1
            #         if column == 4:
            #             column = 0
        self.showForm()

    def clearWindow(self):
        self.form.clear()
        for widget in self.appliViewFrame.winfo_children():
            widget.destroy()

    def reopenView(self):
        self.clearWindow()
        self.form.clear()
        self.openModifyWindow()
        

    def showForm(self):
        """Resets the application view frame and start displaying the form in it
        """
        for widget in self.appliViewFrame.winfo_children():
            try:
                widget.destroy()
            except tk.TclError:
                pass
        self.form.constructView(self.appliViewFrame)

    def completeInsertWindow(self):
        """
        Add the button for an insert window.
            -Insert button that validate the form with the insert function.
        """
        pan = self.form.addFormPanel()
        pan.addFormButton("Insert", self.insert)
        for widget in self.appliViewFrame.winfo_children():
            widget.destroy()
        self.form.constructView(self.appliViewFrame)

    def hide(self, reason="manual_hide"):
        """Tells the application treeview to hide this node
        """
        self.appliTw.hide(str(self.controller.getDbId()), reason)

    def unhide(self):
        """Tells the application treeview to unhide this node
        """
        self.appliTw.unhide(self)

    def __str__(self):
        """
        Return the __str__ method of the model
        """
        return str(self.controller.getModelRepr())

    @classmethod
    def DbToTreeviewListId(cls, parent_db_id):
        """Converts a mongo Id to a unique string identifying a list of view elemnt given its parent
        Args:
            parent_db_id: the parent node mongo ID
        Returns:
            A string that should be unique to describe the parent list of viewelement node
        """
        if parent_db_id is None:
            return None
        return str(parent_db_id)

    def getParentId(self):
        """
        Return the id of the parent node in treeview.

        Returns:
            return the model parent id DbToTreeviewListId
        """
        return self.controller.getParentId()

    def getParentNode(self):
        """
        Return the parent node in treeview.
        """
        return self.__class__.DbToTreeviewListId(self.controller.getParentId())
    
    def openInDialog(self,is_insert=False):
        """
        Open a dialog to show the tool information.
        """
        default_title = "Modify" if not is_insert else "Insert"
        dialog = ChildDialogGenericView(self.mainApp, default_title, self, is_insert=is_insert)
        return dialog.rvalue
    
    def updateReceived(self, obj=None, old_obj=None):
        """Called when any view element update is received by notification.
        Resets the node tags according to database and hide it if "hidden" is in tags
        """
        if self.controller.getDbId() is None:
            return
        tags = self.controller.getTags()
        try:
            self.appliTw.item(str(self.controller.getDbId()), tags=tags)
        except TclError:
            pass
        if "hidden" in tags:
            self.hide("tags")

    def insertReceived(self):
        """Called when any view element insert is received by notificaiton
        To be overriden
        """
        pass

    def key(self):
        """Returns a key for sorting this node
        Returns:
            tuple, key to sort
        """
        return tuple([ord(c) for c in str(self.controller.getModelRepr()).lower()])

    @classmethod
    def list_tuple_to_dict(cls, list_of_tuple):
        """Transforms a list of 2-tuple to a dictionnary
        Args:
            list_of_tuple: a 2-tuple with (key, value)
        Returns:
            A dictionnary with all key-values pair inserted
        """
        ret = dict()
        for key, value in list_of_tuple:
            ret[key] = value
        return ret



"""This class pop a fix view form in a subdialog"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from PIL import ImageTk, Image
from pollenisatorgui.core.application.dialogs.ChildDialogFixView import ChildDialogFixView
import pollenisatorgui.core.components.utils as utils


class ChildDialogFixes:
    """
    Open a child dialog of a tkinter application to answer a question.
    """
    def __init__(self, parent, defectView=None):
        """
        Open a child dialog of a tkinter application to edit fixes settings.

        Args:
            parent: the tkinter parent view to use for this window construction.
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.title("Edit defect fixes")
        self.app.resizable(True, True)
        self.app.bind("<Escape>", self.cancel)
        self.rvalue = None
        self.lastMovedTo = None
        self.pane_base_height = 31
        self.defectView = defectView
        self.fixes = defectView.controller.getData().get("fixes", [])
        appFrame = CTkFrame(self.app)
        
        self.parent = None
        self.initUI(appFrame)
        ok_button = CTkButton(appFrame, text="OK")
        ok_button.pack(side="right", padx=5, pady=10)
        ok_button.bind('<Button-1>', self.okCallback)
        cancel_button = CTkButton(appFrame, text="Cancel", 
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        cancel_button.pack(side="right", padx=5, pady=10)
        cancel_button.bind('<Button-1>', self.cancel)
        appFrame.pack(fill=tk.BOTH, ipady=10, ipadx=10, expand=True)

        self.app.transient(parent)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.grab_set()
            self.app.focus_force()
            self.app.lift()
        except tk.TclError:
            pass

    def cancel(self, _event=None):
        """called when canceling the window.
        Close the window and set rvalue to False
        Args:
            _event: Not used but mandatory"""
        self.rvalue = None
        self.app.destroy()

    def okCallback(self, _event=None):
        """called when pressing the validating button
        Close the window if the form is valid.
        Set rvalue to True and perform the defect update/insert if validated.
        Args:
            _event: Not used but mandatory"""
        
        self.rvalue = self.fixes
        self.app.destroy()

    def initUI(self, parent):
        """
        Initialize window and widgets.
        """
        if self.parent is not None:  # Already initialized
            self.reset()
            self.fillWithFixes()
            return
        self.parent = parent
        ###Fixes TABLE ###
        self.rowHeight = 20
        self.style = ttk.Style()
        self.style.configure('Report.Treeview', rowheight=self.rowHeight)
       
        # FIXES TREEVW	
        fixesLabelFrame = ttk.LabelFrame(parent, text="Fixes table")	
        self.paned = tk.PanedWindow(fixesLabelFrame, orient=tk.VERTICAL, height=400)
        self.frameTw = CTkFrame(self.paned)
        self.treevw = ttk.Treeview(self.frameTw, style='Report.Treeview', height=0)
        self.treevw['columns'] = ('execution', 'gain')
        self.treevw.heading("#0", text='Title', anchor=tk.W)
        self.treevw.column("#0", anchor=tk.W, width=150)
        self.treevw.heading('execution', text='Execution')
        self.treevw.column('execution', anchor='center', width=40)
        self.treevw.heading('gain', text='Gain')
        self.treevw.column('gain', anchor='center', width=40)
        
        self.treevw.bind("<Double-Button-1>", self.OnDoubleClick)
        self.treevw.bind("<Delete>", self.deleteSelectedItem)
        self.treevw.bind("<Alt-Down>",self.bDown)
        self.treevw.bind("<Alt-Up>",self.bUp)
        self.treevw.bind("<ButtonPress-1>",self.dragStart)
        self.treevw.bind("<ButtonRelease-1>",self.dragRelease, add='+')
        self.treevw.bind("<B1-Motion>",self.dragMove, add='+')
        self.treevw.grid(row=0, column=0, sticky=tk.NSEW)
        scbVSel = CTkScrollbar(self.frameTw,
                                orientation=tk.VERTICAL,
                                command=self.treevw.yview)
        self.treevw.configure(yscrollcommand=scbVSel.set)
        scbVSel.grid(row=0, column=1, sticky=tk.NS)
        self.frameTw.pack(side=tk.TOP, fill=tk.BOTH, padx=5, pady=10)
        self.frameTw.columnconfigure(0, weight=1)
        self.frameTw.rowconfigure(0, weight=1)
        ### OFFICE EXPORT FRAME ###
        belowFrame = CTkFrame(self.paned)
        frameBtn = CTkFrame(belowFrame)
        self.buttonUpImage = CTkImage(Image.open(utils.getIcon('up-arrow.png')))
        self.buttonDownImage = CTkImage(Image.open(utils.getIcon('down-arrow.png')))
        # use self.buttonPhoto
        btn_down = CTkButton(frameBtn, image=self.buttonDownImage, text="",command=self.bDown)
        btn_down.pack(side="left", anchor="center")
        btn_up = CTkButton(frameBtn, image=self.buttonUpImage, text="", command=self.bUp)
        btn_up.pack(side="left", anchor="center")
        btn_delFix = CTkButton(
            frameBtn, text="Remove selection", command=self.deleteSelectedItem)
        btn_delFix.pack(side=tk.RIGHT, padx=5)
        btn_addFix = CTkButton(
            frameBtn, text="Add a fix", command=self.addFixCallback)
        btn_addFix.pack(side=tk.RIGHT, padx=5)
        frameBtn.pack(side=tk.TOP, pady=5)
        belowFrame.pack(side=tk.TOP, fill=tk.BOTH)
        self.paned.add(self.frameTw)
        self.paned.add(belowFrame)
        self.paned.pack(fill=tk.BOTH, expand=1)
        fixesLabelFrame.pack(side=tk.TOP, fill=tk.BOTH, pady=10)	
        self.fillWithFixes()

    def bDown(self, event=None):
        item_iid = self.treevw.selection()[0]
        children = self.treevw.get_children()
        iid_moving = children.index(item_iid)
        try:
            iid_moved_by = children[iid_moving+1]
            self.treevw.move(item_iid, '', iid_moving+1)
            self.fixes[iid_moving+1], self.fixes[iid_moving] = self.fixes[iid_moving], self.fixes[iid_moving+1]
        except IndexError:
            pass
        return "break"


    def bUp(self, event=None):
        item_iid = self.treevw.selection()[0]
        children = self.treevw.get_children()
        iid_moving = children.index(item_iid)
        try:
            iid_moved_by = children[iid_moving-1]
            self.treevw.move(item_iid, '', iid_moving-1)
            self.fixes[iid_moving-1], self.fixes[iid_moving] = self.fixes[iid_moving], self.fixes[iid_moving-1]
        except IndexError:
            pass
        return "break"

    def dragStart(self, event):
        tv = event.widget
        if tv.identify_row(event.y) not in tv.selection():
            tv.selection_set(tv.identify_row(event.y))    
            self.movingSelection = tv.identify_row(event.y)

    def dragRelease(self, event):
        if self.movingSelection is None or self.lastMovedTo is None:
            return
        tv = event.widget
        if tv.identify_row(event.y) in tv.selection():
            self.movingSelection = None
            self.lastMovedTo = None

    def dragMove(self, event):
        tv = event.widget
        rowToMove = tv.identify_row(event.y)
        moveto = tv.index(rowToMove)    
        self.lastMovedTo = rowToMove if rowToMove != self.movingSelection else self.lastMovedTo
        for s in tv.selection():
            moved = tv.index(s) 
            if moved != moveto:
                tv.move(s, '', moveto)
                self.fixes[moved], self.fixes[moveto] = self.fixes[moveto], self.fixes[moved]
                
    def deleteSelectedItem(self, _event=None):
        """
        Remove selected defect from treeview
        Args:
            _event: not used but mandatory
        """
        try:
            selected = self.treevw.selection()[0]
        except IndexError:
            return
        self.treevw.delete(selected)
        ind = [fix["title"] for fix in self.fixes].index(str(selected))
        del self.fixes[ind]
        self.resizeTreeview()

    def reset(self):
        """
        reset defect treeview by deleting every item inside.
        """
        for item in self.treevw.get_children():
            self.treevw.delete(item)
        self.fixes = []
        self.resizeTreeview()

    def addFixCallback(self):
        """Open an insert defect view form in a child window"""
        dialog = ChildDialogFixView(self.parent, None)
        self.parent.wait_window(dialog.app)
        if dialog.rvalue is None:
            return
        self.addFixInTreevw(dialog.rvalue)

    def OnDoubleClick(self, event):
        """
        Callback for double click on treeview.
        Opens a window to update the double clicked defect view.
        Args:
            event: automatically created with the event catch. stores data about line in treeview that was double clicked.
        """
        item = self.treevw.identify("item", event.x, event.y)
        if item is None or item == '':
            return
        ind = [fix["title"] for fix in self.fixes].index(str(item))
        fix = self.fixes[ind]
        dialog = ChildDialogFixView(self.parent, fix)
        self.parent.wait_window(dialog.app)
        if dialog.rvalue is None:
            return
        self.fixes[ind] = dialog.rvalue 
        self.treevw.item(item, text=dialog.rvalue["title"], values=(dialog.rvalue["execution"], dialog.rvalue["gain"]))

    def resizeTreeview(self):
        currentHeight = len(self.treevw.get_children())
        self.treevw.configure(height=currentHeight)
        sx, sy = self.paned.sash_coord(0)
        if sy <= (currentHeight)*self.rowHeight + self.pane_base_height:
            self.paned.paneconfigure(self.frameTw, height=(currentHeight)*self.rowHeight + self.pane_base_height)

    def fillWithFixes(self):	
        """	
        Fetch remarks and fill the remarks table with them.	
        """	
        for fix in self.fixes:
            try:
                self.treevw.insert('', "end", fix["title"], text=fix["title"], values=(fix["execution"], fix["gain"],))
            except:
                pass
        self.resizeTreeview()

    def addFixInTreevw(self, fix):
        try:
            self.treevw.insert('', "end", fix["title"], text=fix["title"], values=(fix["execution"], fix["gain"],))
            self.fixes.append(fix)
        except:
            pass
        self.resizeTreeview()


"""Describe tkinter Label with default common args"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.forms.form import Form


class FormLabel(Form):
    """
    Form field representing a label.
    Default setted values:
        width=500
        if pack : padx = pady = 5, side = left
        if grid: row = column = 0 sticky = "East"
    """
    def __init__(self, name, text, **kwargs):
        """
        Constructor for a form label
        Args:
            name: the label name.
            text: the text showed by the label.
            kwargs: same keyword args as you would give to CTkLabel
        """
        super().__init__(name)
        self.text = text
        self.kwargs = kwargs

    def constructView(self, parent):
        """
        Create the label view inside the parent view given

        Args:
            parent: parent FormPanel.
        """
        if self.text == "":
            self.text = self.name + " : "
        width=self.getKw("width", None)
        kwargs = {}
        if width is not None:
            kwargs["width"] = int(width)
        bold = self.getKw("bold", False)
        image = self.getKw("image", None)
        compound = self.getKw("compound", None)
        self.lbl = CTkLabel(parent.panel, text=self.text, justify=tk.LEFT, fg_color=self.getKw("fg_color", None), image=image, compound=compound, **kwargs)
        if bold:
            self.lbl.cget("font").configure(weight="bold")
        self.lbl.cget("font").configure(size=self.getKw("font_size",12))
        try:
            if parent.gridLayout:
        
                self.lbl.grid(column=self.getKw("column", 0), row=self.getKw("row", 0), sticky=self.getKw("sticky", tk.E) , padx=self.getKw("padx", 5), pady=self.getKw("pady", 5), **self.kwargs)
            else:
                self.lbl.pack(side=self.getKw("side", "left"), anchor=self.getKw("anchor", tk.CENTER), padx=self.getKw("padx", 10), pady=self.getKw("pady", 5), **self.kwargs)
        except tk.TclError as e:
            raise tk.TclError("Error while creating label '" + self.name + "' : " + str(e))

    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return the label text.
        """
        return self.text
    
    def setValue(self, newval):
        """nothing to set so overwrite
        Args:
            _newval: not used"""
        self.text = newval
        self.lbl.configure(text=self.text)
        return


"""Hold functions to interact form the scan tab in the notebook"""
from pollenisatorgui.core.application.dialogs.ChildDialogAutoScanParams import ChildDialogAutoScanParams
from pollenisatorgui.core.application.dialogs.ChildDialogScanHistory import ChildDialogScanHistory
from pollenisatorgui.core.application.dialogs.ChildDialogToolsInstalled import ChildDialogToolsInstalled
from pollenisatorgui.core.application.scrollableframexplateform import ScrollableFrameXPlateform
from pollenisatorgui.core.application.scrollabletreeview import ScrollableTreeview
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.datamanager import DataManager
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import threading
import time
from pollenisatorgui.core.components.scanworker import ScanWorker
from pollenisatorgui.core.controllers.toolcontroller import ToolController
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.models.tool import Tool
from pollenisatorgui.core.application.dialogs.ChildDialogFileParser import ChildDialogFileParser
from pollenisatorgui.core.application.dialogs.ChildDialogProgress import ChildDialogProgress
from pollenisatorgui.core.application.dialogs.ChildDialogCombo import ChildDialogCombo
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
from pollenisatorgui.core.models.checkinstance import CheckInstance
from PIL import Image
import pollenisatorgui.core.components.utils as utils
import os
import docker
from bson import ObjectId
from pollenisatorgui.core.components.logger_config import logger

import shutil


def start_docker(dialog, force_reinstall):
    dialog.update(msg="Building worker docker could take a while (1~10 minutes depending on internet connection speed)...")
    try:
        client = docker.from_env()
        clientAPI = docker.APIClient()
    except Exception as e:
        dialog.destroy()
        tk.messagebox.showerror("Unable to launch docker", e)
        return
    try:
        log_generator = clientAPI.pull("algosecure/pollenisator-worker:latest",stream=True,decode=True)
        change_max = None
        for byte_log in log_generator:
            log_line = byte_log["status"].strip()
            dialog.update(log=log_line+"\n")
    except docker.errors.APIError as e:
        dialog.destroy()
        tk.messagebox.showerror("APIError docker error", "Pulling error:\n"+str(e))
        return
    image = client.images.list("algosecure/pollenisator-worker")
    if len(image) == 0:
        tk.messagebox.showerror("Pulling docker failed", "The docker pull command failed, try to install manually...")
        return
    dialog.update(2, msg="Starting worker docker ...")
    clientCfg = utils.loadClientConfig()
    if clientCfg["host"] == "localhost" or clientCfg["host"] == "127.0.0.1":
        network_mode = "host"
    else:
        network_mode = None
    container = client.containers.run(image=image[0], network_mode=network_mode, volumes={os.path.join(utils.getConfigFolder()):{'bind':'/root/.config/pollenisator-gui/', 'mode':'rw'}}, detach=True)
    dialog.update(3, msg="Checking if worker is running")
    print(container.id)
    if container.logs() != b"":
        print(container.logs())
    dialog.destroy()
    
class ScanManager:
    """Scan model class"""

    def __init__(self, mainApp, nbk, linkedTreeview, pentestToScan, settings):
        self.pentestToScan = pentestToScan
        self.mainApp = mainApp
        self.nbk = nbk
        self.settings = settings
        self.btn_autoscan = None
        self.workers = None
        self.btn_docker_worker = None
        self.parent = None
        self.sio = None
        self.workerTv = None
        self.linkTw = linkedTreeview
        self.scan_worker = ScanWorker(self.settings)
        self.tool_icon = tk.PhotoImage(file=utils.getIcon("tool.png"))
        self.nok_icon = tk.PhotoImage(file=utils.getIcon("cross.png"))
        self.waiting_icon = tk.PhotoImage(file=utils.getIcon("waiting.png"))
        self.ok_icon = tk.PhotoImage(file=utils.getIcon("done_tool.png"))
        self.running_icon = tk.PhotoImage(file=utils.getIcon("running.png"))
        DataManager.getInstance().attach(self)

    def startAutoscan(self):
        """Start an automatic scan. Will try to launch all undone tools."""
        apiclient = APIClient.getInstance()
        if len(self.workerTv.get_children()) == 0:
            if not self.ask_start_worker():
                return
        workers = apiclient.getWorkers({"pentest":apiclient.getCurrentPentest()})
        workers = [w for w in workers]
        if len(workers) == 0:
            tk.messagebox.showwarning("No selected worker found", "A worker exist but is not registered for this pentest. You might want to register it by double clicking on it or using the Use button.")
            return False
        dialog = ChildDialogAutoScanParams(self.parent)
        dialog.wait_window()
        params = dialog.rvalue
        if params is None:
            return
        if self.settings.db_settings.get("include_all_domains", False):
            answer = tk.messagebox.askyesno(
                "Autoscan warning", "The current settings will add every domain found in attack's scope. Are you sure ?")
            if not answer:
                return False
        self.btn_autoscan.configure(text="Stop Scanning", command=self.stopAutoscan)
        apiclient.sendStartAutoScan(command_iids=params["commands"], autoqueue=params["autoqueue"])
        return True
    
    def stop(self):
        """Stop an automatic scan. Will try to stop running tools."""
        apiclient = APIClient.getInstance()
        apiclient.sendStopAutoScan()
        logger.debug('Ask stop autoscan')
        return True

    def refreshRunningScans(self):
        running_scans = list(Tool.fetchObjects({"status":"running"}))
        checks = CheckInstance.fetchObjects([str(ObjectId(running_scan.check_iid)) for running_scan in running_scans if running_scan.check_iid != ""])
        mapping = {}
        for check in checks:
            mapping[str(check._id)] = check
        try:
            for children in self.scanTv.get_children():
                self.scanTv.delete(children)
        except tk.TclError:
            pass
        except RuntimeError:
            return
        for running_scan in running_scans:
            check = mapping.get(str(running_scan.check_iid), None)
            group_name = "" if check is None else check.check_m.title
            try:
                self.scanTv.insert('','end', running_scan.getId(), text=group_name, values=(running_scan.name, running_scan.dated), image=self.running_icon)
            except tk.TclError:
                pass

    def refreshQueuedScans(self):
        apiclient = APIClient.getInstance()
        queued_scans = apiclient.getQueue()
        if queued_scans is None:
            return 
        checks = CheckInstance.fetchObjects([str(ObjectId(queued_scan["check_iid"])) for queued_scan in queued_scans if queued_scan["check_iid"] != ""])
        mapping = {}
        for check in checks:
            mapping[str(check._id)] = check
        try:
            for children in self.queueTv.get_children():
                self.queueTv.delete(children)
        except tk.TclError:
            pass
        except RuntimeError:
            return
        for queued_scan in queued_scans:
            check = mapping.get(str(queued_scan["check_iid"]), None)
            group_name = "" if check is None else check.check_m.title
            try:
                self.queueTv.insert('','end', str(queued_scan["_id"]), text=group_name, values=(queued_scan["name"], queued_scan["text"]), image=self.waiting_icon, auto_update_pagination=False)
            except tk.TclError:
                pass
        self.queueTv.setPaginationPanel()

    def is_worker_valid_for_pentest(self, worker_data):
        apiclient = APIClient.getInstance()
        return worker_data.get("pentest", "") == apiclient.getCurrentPentest()


    def refreshWorkers(self):
        apiclient = APIClient.getInstance()
        workers = apiclient.getWorkers()
        self.workers = workers
        try:
            for children in self.workerTv.get_children():
                self.workerTv.delete(children)
        except:
            pass
        registeredCommands = set()
        for worker in workers:
            workername = worker["name"]
            try:
                if self.is_worker_valid_for_pentest(worker):
                    worker_node = self.workerTv.insert(
                        '', 'end', workername, text=workername, image=self.ok_icon)
                else:
                    worker_node = self.workerTv.insert(
                        '', 'end', workername, text=workername, image=self.nok_icon)
            except tk.TclError as err:
                try:
                    worker_node = self.workerTv.item(workername)["text"]
                except Exception as e:
                    print(str(err)+" occured")
                    print("Then:"+str(e))
            except RuntimeError:
                pass
        return len(workers)
    
    def is_already_queued(self, tool_iid):
        try:
            children = self.queueTv.get_children()
            return tool_iid in children
        except tk.TclError:
            return False


    def _check_worker(self):
        foundValidWorker = False
        
        if self.workers:
            for worker in self.workers:
                if self.is_worker_valid_for_pentest(worker):
                    foundValidWorker = True
                    break
        if not foundValidWorker:
            return False, "No worker available yet."
        return True, ""
    
    def is_ready_to_queue(self, tool_iid=None):
        apiclient = APIClient.getInstance()
        res, msg = self._check_worker()
        if not res:
            return res, msg
        if not apiclient.getAutoScanStatus():
            return False, "No autoscan running."
        if tool_iid is not None:
            if self.is_already_queued(tool_iid):
                return False, "Tool already queued."
        return True, ""
    
    def is_ready_to_run_tasks(self):
        res, msg = self._check_worker()
        if not res:
            return res, msg
        return True, ""

    def slider_event(self, event):
        val = self.autoscan_slider.get()
        val = int(val)
        self.autoscan_threads_lbl.configure(text=str(val) + " thread"+("s" if val > 1 else ""))
        self.settings.setPentestSetting("autoscan_threads", val)

    def update_thread_label(self, event):
        val = self.autoscan_slider.get()
        val = int(round(val)) 
        self.autoscan_threads_lbl.configure(text=str(val) + " thread"+("s" if val > 1 else ""))

    def refreshUI(self):
        """Reload informations and renew widgets"""
        apiclient = APIClient.getInstance()
        self.refreshRunningScans()
        self.refreshQueuedScans()
        self.settings._reloadDbSettings()
        v = int(self.settings.db_settings.get("autoscan_threads", 4))
        self.autoscan_slider.set(v)
        self.update_thread_label(v)
        
        nb_workers = self.refreshWorkers()
        
        if self.btn_autoscan is None:
            if apiclient.getAutoScanStatus():
                self.btn_autoscan = CTkButton(
                    self.parent, text="Stop Scanning", image=self.image_auto, command=self.stopAutoscan, fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
            else:
                self.btn_autoscan = CTkButton(
                    self.parent, text="Start Scanning", command=self.startAutoscan)
        
        if nb_workers == 0:
            self.ask_start_worker()
        logger.debug('Refresh scan manager UI')

    def ask_start_worker(self):
        options = ["Use this computer", "Run a preconfigured Docker on server"]
        options.append("Run a preconfigured Docker locally")
        options.append("Cancel")
        dialog = ChildDialogQuestion(self.parent, "Register worker ?", "There is no running scanning clients. What do you want to do ?", options)
        self.parent.wait_window(dialog.app)
        if dialog.rvalue is not None:
            rep = options.index(dialog.rvalue)
            if rep == 1:
                self.runWorkerOnServer()
            elif rep == 0:
                self.registerAsWorker()
            elif rep == 2:
                self.launchDockerWorker()
            return True
        return False

    def initUI(self, parent):
        """Create widgets and initialize them
        Args:
            parent: the parent tkinter widget container."""
        if self.parent is not None:
            self.refreshUI()
            return
        apiclient = APIClient.getInstance()
        self.parent = parent
        parentScrollableFrame = ScrollableFrameXPlateform(self.parent)
        parentFrame = ttk.Frame(parentScrollableFrame)
        parentFrame.columnconfigure(0, weight=1)
        parentFrame.columnconfigure(1, weight=1)
        parentFrame.rowconfigure(0, weight=1)
        parentFrame.configure(onfiledrop=self.dropFile)
        ### WORKER TREEVIEW : Which worker knows which commands
        workerScanFrame = CTkFrame(parentFrame)
        workerFrame = CTkFrame(workerScanFrame)
        workerFrame.columnconfigure(0, weight=1)
        self.workerTv = ttk.Treeview(workerFrame, height=3)
        self.workerTv['columns'] = ('workers')
        self.workerTv.heading("#0", text='Workers', anchor=tk.W)
        self.workerTv.column("#0", anchor=tk.W)
        self.workerTv.grid(row=0, column=0, sticky=tk.NSEW, padx=10, pady=10)
        self.workerTv.bind("<Double-Button-1>", self.OnWorkerDoubleClick)
        self.workerTv.bind("<Delete>", self.OnWorkerDelete)
        btn_pane = FormPanel(row=0, column=1, sticky=tk.W+tk.E, grid=True)
        pane = btn_pane.addFormPanel(row=0,column=0)
        pane.addFormButton("Use/Stop using selected worker", callback=self.setWorkerInclusion, side=tk.LEFT)
        pane.addFormHelper("Give / Remove the right for a worker to work for the current pentest", side=tk.LEFT)
        pane = btn_pane.addFormPanel(row=1,column=0)
        pane.addFormButton("Start remote worker",  callback=self.runWorkerOnServer, side=tk.LEFT)
        pane.addFormHelper("Start a docker worker on the remote server", side=tk.LEFT)
        
        pane = btn_pane.addFormPanel(row=2,column=0)
        pane.addFormButton("Start worker locally", callback=self.launchDockerWorker, side=tk.LEFT)
        pane.addFormHelper("Require Docker, pull and run the worker docker locally", side=tk.LEFT)
        pane = btn_pane.addFormPanel(row=3,column=0)
        pane.addFormButton("Use this computer", callback=self.registerAsWorker, side=tk.LEFT)
        pane.addFormHelper("Use this computer as a worker", side=tk.LEFT)
        btn_pane.constructView(workerFrame)
        workerFrame.pack(side=tk.TOP, pady=5)
        self.image_auto = CTkImage(Image.open(utils.getIcon("auto.png")))
        self.image_import = CTkImage(Image.open(utils.getIcon("import.png")))
        if apiclient.getAutoScanStatus():
            self.btn_autoscan = CTkButton(
                workerScanFrame, text="Stop Scanning", image=self.image_auto, command=self.stopAutoscan)
            self.btn_autoscan.pack()
        else:
            self.btn_autoscan = CTkButton(
                workerScanFrame, text="Start Scanning", image=self.image_auto, command=self.startAutoscan)
            self.btn_autoscan.pack()
        frame_settings = CTkFrame(workerScanFrame)
        self.autoscan_slider = CTkSlider(frame_settings, from_=1, to=10, number_of_steps=10, command=self.update_thread_label)
        self.autoscan_slider.bind("<ButtonRelease-1>", self.slider_event)
        self.autoscan_slider.pack(side=tk.LEFT, padx=5)
        threads = int(self.settings.db_settings.get("autoscan_threads", 4))
        self.autoscan_slider.set(threads)
        self.autoscan_threads_lbl = CTkLabel(frame_settings, text=str(threads)+" thread"+("s" if threads > 1 else ""))
        self.autoscan_threads_lbl.pack(side=tk.LEFT, padx=5)
        frame_settings.pack()
        workerScanFrame.grid(row=0, column=0, sticky=tk.NSEW)
        importScanFrame = CTkFrame(parentFrame)
        sep = ttk.Separator(importScanFrame, orient="vertical")
        sep.pack(side="left", fill="y" , padx=10, pady=10)
        btn_parse_scans = CTkButton(
            importScanFrame, text="Parse existing files", image=self.image_import, command=self.parseFiles)
        btn_parse_scans.place(relx=0.5, rely=0.5, anchor=CENTER)
        info = CTkLabel(importScanFrame, text="You can also drop your files / folder here")
        info.pack(side=tk.BOTTOM, pady=10)
        importScanFrame.grid(row=0, column=1, padx=10, sticky=tk.NSEW)
        ####### RUNNING SCANS
        scansInfoFrame = CTkFrame(parentFrame)
        
        self.scanTv = ttk.Treeview(scansInfoFrame)
        self.scanTv['columns'] = ('Tool', 'Started at')
        self.scanTv.heading("#0", text='Running scans', anchor=tk.W)
        self.scanTv.column("#0", anchor=tk.W)
        self.scanTv.pack(side=tk.TOP, padx=10, pady=10, fill=tk.X)
        self.scanTv.bind("<Double-Button-1>", self.OnDoubleClick)
        panelActions = CTkFrame(scansInfoFrame)
        refresh_button = CTkButton(panelActions, text="Refresh scans", command=self.refreshRunningScans)
        refresh_button.pack(side=tk.LEFT, pady=2, padx=10)
        self.show_history = CTkButton(panelActions, text="Show history", command=self.showHistory)
        self.show_history.pack(side=tk.LEFT, pady=2, padx=10)
        panelActions.pack(side=tk.BOTTOM, pady=2, padx=10)
        scansInfoFrame.grid(row=1, column=0, columnspan=2, sticky=tk.NSEW)
        ###### QUEUED
        queuedScansFrame = CTkFrame(parentFrame)
        
        self.queueTv = ScrollableTreeview(queuedScansFrame,  ("Check type", 'Tool', 'options'))
        self.queueTv.pack(side=tk.TOP, padx=10, pady=10, fill=tk.X)
        self.queueTv.bind("<Delete>", self.remove__selected_tasks)
        clear_queue_btn = CTkButton(queuedScansFrame, text="Clear queue", command=self.clear_queue)
        clear_queue_btn.pack(side=tk.BOTTOM, pady=2, padx=10)
        queuedScansFrame.grid(row=2, column=0, columnspan=2, sticky=tk.NSEW)

        ######
        parentFrame.pack(expand=1, fill=tk.BOTH)
        parentScrollableFrame.pack(expand=1, fill=tk.BOTH)

    def clear_queue(self, event=None):
        apiclient = APIClient.getInstance()
        apiclient.clear_queue()

    def remove__selected_tasks(self, _event=None):
        items = self.queueTv.selection()
        if items:
            apiclient = APIClient.getInstance()
            apiclient.sendRemoveTasks(items)

    def showHistory(self, event=None):
        dialog = ChildDialogScanHistory(self.parent)

    def dropFile(self, event):
        # This function is called, when stuff is dropped into a widget
        data = utils.drop_file_event_parser(event)
        self.parseFiles(data)

    def OnDoubleClick(self, event):
        """Callback for a double click on ongoing scan tool treeview. Open the clicked tool in main view and focus on it.
        Args:
            event: Automatically filled when event is triggered. Holds info about which line was double clicked
        """
        if self.scanTv is not None:
            tv = event.widget
            item = tv.identify("item", event.x, event.y)
        self.openInTerminalView(item)
        self.openInMainView(item)

    def openInMainView(self, item):
        self.nbk.select("Main View")
        self.mainApp.search("id == \""+str(item)+"\"")
    
    def openInTerminalView(self, item):
        datamanager = DataManager.getInstance()
        tool = datamanager.get("tools", str(item))
        if tool:
            tool_controller = ToolController(tool)
            self.mainApp.open_any_terminal(str(tool.check_iid)+"|"+str(tool_controller.getDbId()), tool_controller.getDetailedString(), tool_controller, self)
    
    def stopAutoscan(self):
        """
        Stop an automatic scan. Will terminate celery running tasks.
        """
        try:
            if self.btn_autoscan is not None:
                self.btn_autoscan.configure(
                    text="Start Scanning", command=self.startAutoscan)
        except tk.TclError:
            pass
        print("Stopping auto... ")
        apiclient = APIClient.getInstance()
        apiclient.sendStopAutoScan()
        logger.debug('Ask stop autoscan from UI')

    def parseFiles(self, default_path=""):
        """
        Ask user to import existing files to import.
        """
        dialog = ChildDialogFileParser(self.mainApp, default_path)
        self.parent.wait_window(dialog.app)

    def update_received(self, dataManager, notif, obj, old_obj):
        """
        Reload UI when notified
        """
        if notif["db"] == "pollenisator":
            if notif["collection"] == "workers":
                if self.workerTv is not None:
                    self.refreshWorkers() 
        # elif notif["collection"] == "tools" and notif["action"] == "update":
        #     if dataManager.currentPentest == notif["db"]:
        #         self.refreshRunningScans()
        #         self.refreshQueuedScans()
        

    def OnWorkerDoubleClick(self, event):
        """Callback for treeview double click.
        If a link treeview is defined, open mainview and focus on the item with same iid clicked.
        Args:
            event: used to identified which link was clicked. Auto filled
        """
        if self.workerTv is not None:
            if event:
                tv = event.widget
                item = tv.identify("item", event.x, event.y)
            parent = self.workerTv.parent(item)
            self.setUseForPentest(item)
    
    def setWorkerInclusion(self, _event=None):
        items = self.workerTv.selection()
        for item in items:
            if "|" not in self.workerTv.item(item)["text"]: # exclude tools and keep worker nodes
                self.setUseForPentest(item)

    def setUseForPentest(self, worker_hostname):
        apiclient = APIClient.getInstance()
        worker = apiclient.getWorker({"name":worker_hostname})
        if worker is not None:
            isIncluded = apiclient.getCurrentPentest() == worker.get("pentest", "")
            apiclient.setWorkerInclusion(worker_hostname, not (isIncluded))

    def getToolProgress(self, toolId):
        return self.scan_worker.getToolProgress(toolId)

    def stopTask(self, toolId):
        return self.scan_worker.stopTask(toolId)

    def launchTask(self, toolModel, checks=True, worker="", infos={}):
        return self.scan_worker.launchTask(toolModel, checks=checks, worker=worker, infos=infos)


    def OnWorkerDelete(self, event):
        """Callback for a delete key press on a worker.
        Force deletion of worker
        Args:
            event: Auto filled
        """
        apiclient = APIClient.getInstance()
        dialog = ChildDialogProgress(self.parent, "Docker delete", "Waiting for worker to stop", progress_mode="indeterminate")
        dialog.show()
        apiclient.deleteWorker(self.workerTv.selection()[0]) 
        dialog.destroy()
    

    def launchDockerWorker(self, event=None):
        dialog = ChildDialogProgress(self.parent, "Starting worker docker", "Initialisation ...",  progress_mode="indeterminate", show_logs=True)
        dialog.show(4)
        x = threading.Thread(target=start_docker, args=(dialog, True))
        x.start()
        return x


    def runWorkerOnServer(self, _event=None):
        apiclient = APIClient.getInstance()
        docker_name = apiclient.getDockerForPentest(apiclient.getCurrentPentest())
        dialog = ChildDialogProgress(self.parent, "Start docker", "Waiting for docker to boot 0/4", progress_mode="indeterminate")
        dialog.show()
        nb_try = 0
        max_try = 3
        while docker_name not in self.workerTv.get_children() and nb_try < max_try:
            dialog.update(msg=f"Waiting for docker to boot {nb_try+1}/{max_try+1}")
            time.sleep(3)
            nb_try += 1
        dialog.destroy()
        if docker_name not in self.workerTv.get_children():
            return False, "Worker did not boot in time, cannot add commands to wave"
        return True, ""

    def onClosing(self):
        logger.debug("Scan manager on closing state.")
        try:
            client = docker.from_env()
            containers = client.containers.list()
            for container in containers:
                if container.image.tags[0].startswith("algosecure/pollenisator-worker"):
                    
                    dialog = ChildDialogProgress(self.parent, "Stopping docker", "Waiting for docker to stop", progress_mode="indeterminate")
                    dialog.show()
                    logger.debug("Stopping running worker....")
                    container.stop()
                    container.remove()
                    logger.debug("done.")
                    dialog.destroy()
        except Exception as e:
            pass
        self.scan_worker.onClosing()
        apiclient = APIClient.getInstance()
        apiclient.deleteWorker(apiclient.getUser()) 
        if self.sio is not None:
            self.sio.disconnect()

    

    def is_local_launched(self, toolId):
        return self.scan_worker.is_local_launched(toolId)

    def registerAsWorker(self, _event=None):
        prog = ChildDialogProgress(self.parent, "Testing local tools", "Testing which tools will be available...", progress_mode="indeterminate")
        prog.show()
        prog.update()
        results = self.mainApp.testLocalTools()
        if len(results["failures"]) > 0:
            dialog = ChildDialogToolsInstalled(results)
            if dialog.rvalue is not None:
                self.settings.local_settings["my_commands"] = dialog.rvalue
                self.settings.saveLocalSettings()
        prog.destroy()

        self.settings.reloadLocalSettings()
        apiclient = APIClient.getInstance()
        name = apiclient.getUser()
        self.scan_worker.connect(name)
        dialog = ChildDialogProgress(self.parent, "Registering", "Waiting for register 0/4", progress_mode="indeterminate")
        dialog.show()
        nb_try = 0
        max_try = 3
        while name not in self.workerTv.get_children() and nb_try < max_try:
            dialog.update(msg=f"Waiting for registering {nb_try+1}/{max_try+1}")
            time.sleep(2)
            nb_try += 1
        dialog.destroy()
        if name not in self.workerTv.get_children():
            return False, "Worker did not boot in time, cannot add commands to wave"
        apiclient.setWorkerInclusion(name, True)



import multiprocessing
import os
import pty
import select
import shutil
import signal
import subprocess
import time
import tkinter as tk
from tkinter import ttk
from customtkinter import *
import libtmux
from pollenisatorgui.core.application.pseudotermframe import PseudoTermFrame
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.components.utils import getMainDir, getIcon, craftMenuWithStyle, read_and_forward_pty_output
from pollenisatorgui.core.components.logger_config import logger
from PIL import Image, ImageTk



# def read_and_forward_pty_output(fd, queue=None, queueResponse=None):
#     max_read_bytes = 1024 * 20
#     try:
#         while True:
#             time.sleep(0.01)
#             if fd:
#                 timeout_sec = 0
#                 (data_ready, _, _) = select.select([fd], [], [], timeout_sec)
#                 if not queue.empty():
#                     command = queue.get()
#                     os.write(fd, command.encode())
#                 if data_ready:
#                     output = os.read(fd, max_read_bytes).decode(
#                         errors="ignore"
#                     )
#                     queueResponse.put(output.replace("\r",""))
#     except Exception as e:
#         pass

def killThisProc(proc):
    try:
        logger.error("TERMINAL : Killing process terminal")
        time.sleep(1) # HACK to avoid xterm crash ✨ black magic ✨
        os.kill(proc.pid, signal.SIGTERM)
        
    except Exception as e:
        pass

class TerminalsWidget(CTkFrame):
    cachedClassIcon = None
    cachedPseudoTerminalClassIcon = None
    icon = "terminal_small.png"
    iconPseudoTerminal = "pseudo_terminal.png"
    def __init__(self, parent,  mainApp, **kwargs):
        super().__init__(parent,  **kwargs)
        self.parent = parent
        
        self.child_pid = None
        self.pseudoTermFrames = {}
        self.terminalFrames = {}
        self.proc = None
        self.fd = None
        self.mainApp = mainApp
        self.mainApp.subscribe_notification("notif_terminal", self.notif_terminal)
        self.inited = False
        self.opened = None
        self.initUI()

    def initUI(self):
        self.panedwindow = tk.PanedWindow(self, orient=tk.HORIZONTAL, height=200) 
        s = ttk.Style()
        s.configure('Terminal.Treeview.Item', indicatorsize=0)
        self.terminalTv = ttk.Treeview(self.panedwindow, show="tree", selectmode="browse", style="Terminal.Treeview")
        self.contextualMenu = craftMenuWithStyle(self.parent)
        self.contextualMenu.add_command(
            label="Close term", command=self.closeTerm)
        self.contextualMenu.add_command(
            label="Close done terms", command=self.closeDoneTerms)
        self.contextualMenu.add_command(
            label="Close all terms", command=self.closeAllTerms)
        self.contextualMenu.add_command(
            label="(Exit)", command= lambda: 0)# do nothing
        self.terminalTv.bind("<Button-3>", self.doPopup)
        self.terminalTv.column('#0', minwidth=50, width=80, stretch=tk.YES)
        self.proxyFrame = tk.Frame(self)
        self.pseudoTerminalFrame = tk.Frame(self.proxyFrame)
        self.terminalFrame = tk.Frame(self.proxyFrame)
        self.terminalFrame.pack(fill=tk.BOTH, expand=True)
        self.proxyFrame.pack(fill=tk.BOTH, expand=True)
        self.terminalTv.pack(fill=tk.Y, expand=True, padx=0,ipadx=0)
        self.terminalTv.bind("<<TreeviewSelect>>", self.onTreeviewSelect)
        self.terminalTv.tag_configure("notified", background="red")
        self.panedwindow.add(self.terminalTv)
        self.panedwindow.add(self.proxyFrame)
        self.panedwindow.pack(fill=tk.BOTH, expand=True)

    

    def doPopup(self, event):
        """Open the popup 
        Args:
            event: filled with the callback, contains data about line clicked
        """
        self.contextualMenu.selection = self.terminalTv.identify(
            "item", event.x, event.y)
        # display the popup menu
        try:
            self.contextualMenu.tk_popup(event.x_root, event.y_root)
        except Exception as e:
            print(e)
        finally:
            # make sure to release the grab (Tk 8.0a1 only)
            self.contextualMenu.grab_release()
        self.contextualMenu.focus_set()
        self.contextualMenu.bind('<FocusOut>', self.popupFocusOut)

    def popupFocusOut(self, _event=None):
        """Called when the contextual menu loses focus. Closes it.
        Args:
            _event: default to None
        """
        self.contextualMenu.unpost()
    
    def onTreeviewSelect(self, event=None):
        selection = self.terminalTv.selection()
        if len(selection) == 1:
            try:
                tags = self.terminalTv.item(str(selection[0]))["tags"]
                if "notified" in tags:
                    self.terminalTv.item(str(selection[0]), tags="notified-read")
            except tk.TclError:
                pass
            self.view_window(str(selection[0]))

    def update_size(self, event=None):
        xterm_width = self.terminalFrame.winfo_width()
        xterm_height = self.terminalFrame.winfo_height()
        if self.child_pid is not None:
            #HACK: only way I could find to resize the window is xdotool.
            xdotool_command = f"xdotool search --class popoxterm windowsize $(xdotool search --class popoxterm) {xterm_width} {xterm_height}"
            subprocess.call(xdotool_command, shell=True)

    def onClosing(self, _signum=None, _frame=None):
        if self.child_pid is not None:
            os.kill(self.child_pid, signal.SIGTERM)
        self.inited = False
        self.terminalTv.delete(*self.terminalTv.get_children())

    @classmethod
    def getIcon(cls):
        if cls.cachedClassIcon is None:
            path = getIcon(cls.icon)
            img = Image.open(path)
            cls.cachedClassIcon = ImageTk.PhotoImage(img)
        return cls.cachedClassIcon
    
    @classmethod
    def getPseudoTerminalIcon(cls):
        if cls.cachedPseudoTerminalClassIcon is None:
            path = getIcon(cls.iconPseudoTerminal)
            img = Image.open(path)
            cls.cachedPseudoTerminalClassIcon = ImageTk.PhotoImage(img)
        return cls.cachedPseudoTerminalClassIcon
    

    def open_terminal(self, iid=None, title="", enable_trap=True):
        if iid is not None:
            self.create_window(iid, title, enable_trap=enable_trap)
        elif not self.inited:
            self.create_terminal()
            self.after(100, self.update_size)
            self.inited = True
        return
    

    def launch_in_terminal(self, iid, commandline, use_pollex=True):
        session = self.get_session()
        settings = Settings()
        settings.reloadLocalSettings()
        if not settings.isTrapCommand() and use_pollex:
            commandline = "pollex "+commandline
        if settings.isTrapCommand() and not use_pollex:
            commandline = commandline
            
        if session is not None:
            window = session.windows.filter(window_name=str(iid))[0]
            self.terminalFrames[iid] = commandline
            window.select_window()
            window.panes[0].send_keys(commandline)


    def open_ro_terminal(self, iid, title, tool_controller, scanManager):
        if iid is not None:
            self.create_ro_window(iid, title, tool_controller, scanManager)
        return
    
    def open_any_terminal(self, iid, title, tool_controller, scanManager):
        if iid is not None:
            if iid in self.terminalFrames:
                self.open_terminal(iid, title)
            else:
                self.open_ro_terminal(iid, title, tool_controller, scanManager)

        return  

    def get_session(self):
        sessions = self.s.sessions.filter(session_name="pollenisator")
        if sessions:
            session = sessions[0]
        else:
            session = None
        return session

    def notif_terminal(self, notification):
        iid = notification.get("iid",{})
        check_iid = iid.get("check_iid", "")
        tool_iid = iid.get("tool_iid")
        iid = str(check_iid) if tool_iid is None else str(check_iid)+"|"+str(tool_iid)
        if self.opened != str(iid):
            try:
                self.terminalTv.item(str(iid), tags="notified")
            except tk.TclError: # item not found, may append if terminal is not opened
                pass
        return

    def create_terminal(self):
        settings = Settings()
        settings.reloadLocalSettings()
        self.terminalFrame.wait_visibility()

        self.wid = self.terminalFrame.winfo_id()
        self.s = libtmux.Server()
        (child_pid, fd) = pty.fork()
        self.child_pid = child_pid
        child_pid
        if child_pid == 0:
            # child process with ✨ black magic ✨
            try:
                logger.debug("Creating terminal...")
                session_name = "pollenisator"
                sessions = self.s.sessions.filter(session_name=session_name)
                if len(sessions) > 0:
                    for session in sessions:
                        session.kill_session()
                wid = self.wid
                logger.debug(f"Populatig window id {wid}")
                config_location = os.path.join(getMainDir(), "config/")
                xterm_conf = os.path.join(config_location, ".Xresources")
                tmux_conf = os.path.join(config_location, ".tmux.conf")
                terminal_conf = os.path.join(config_location, "shell_ressources")
                logger.debug(f"Trying to load xterm conf through xrdp {xterm_conf}")
                subprocess.run("xrdb -load %s" % xterm_conf, shell=True)
                is_there_zsh = os.environ.get("ZSH",None) is not None
                shell_command = settings.local_settings.get("terminal",  os.environ.get("SHELL","zsh" if is_there_zsh else "/bin/bash"))
                logger.debug(f"shell_command found : {shell_command}")
                trap_suffix = ("trap" if settings.isTrapCommand()  else "notrap")
                logger.debug(f"trap_suffix found : {trap_suffix}")
                if os.path.basename(shell_command) == "zsh":
                    terminal_conf = os.path.join(terminal_conf, "zshrc_"+trap_suffix)
                    default_command = f"ZDOTDIR={terminal_conf} {shell_command}"
                else:
                    terminal_conf = os.path.join(terminal_conf, "bash_setupTerminalForPentest_"+trap_suffix+".sh")
                    default_command = f"{shell_command} --rcfile {terminal_conf}"
                logger.debug(f"terminal_conf found : {terminal_conf}")
                logger.debug(f"default_command found : {default_command}")
                tmux_conf_new = tmux_conf+".popo"
                with open(tmux_conf, "r") as f:
                    tmux_conf_content = f.read()
                    with open(tmux_conf_new, "w") as f2:
                        if "/bin" not in shell_command:
                            shell_command = "/bin/"+shell_command
                        tmux_conf_content += f"\nset -g default-shell {shell_command}"
                        tmux_conf_content += f"\nset -g default-command \"{default_command}\"\n"
                        f2.write(tmux_conf_content)
                    tmux_conf = tmux_conf_new
                command = f"xterm -into {wid} -class popoxterm -name popoxterm -e \"tmux -f {tmux_conf} new-session -s {session_name} -n shell\""
                logger.debug(f"Lauching terminal : {command}")
                proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE, shell=True, preexec_fn=os.setsid)
                signal.signal(signal.SIGINT, lambda signum,sigframe: killThisProc(proc))
                signal.signal(signal.SIGTERM, lambda signum,sigframe: killThisProc(proc))
                # Wait for session to pop
                for i in range(3):
                    if len(self.s.sessions.filter(session_name=session_name)) > 0:
                        break
                    time.sleep(0.5*i)
                
                session = self.get_session()
                if session is not None:
                    logger.debug(f"SUCCESS Lauching terminal : session found")
                    for i in range(3):
                        if len(session.windows.filter(window_name="shell")) > 0:
                            break
                        time.sleep(0.5*i)
                    if len(session.windows.filter(window_name="shell")) > 0:
                        window = session.windows.filter(window_name="shell")[0]
                        window.select_window()
                        
            except Exception as e:
                logger.error("TERMINAL exception: "+str(e))
                sys.exit(0)
            try:
                proc._killed = False
                stdout, stderr = proc.communicate() # wait for ending
                logger.error("TERMINAL ended: "+str(stdout)+"///"+str(stderr))
                sys.exit(0)
            except Exception as e:
                logger.error("TERMINAL end exception: "+str(e))
                sys.exit(0)
                
        else:
            
            queue = multiprocessing.Queue()
            queueResponse = multiprocessing.Queue()
            p = multiprocessing.Process(target=read_and_forward_pty_output, args=[fd, child_pid, queue, queueResponse, False])
            p.start()
            self.terminalFrame.bind("<Configure>", self.update_size)
            self.update_size()
            self.terminalTv.insert("", "end", "shell", text="shell", image=TerminalsWidget.getIcon())
            self.terminalTv.selection_set("shell")
    
    def view_window(self, iid):
        if iid.endswith("|ro"):
            if self.terminalFrame.winfo_viewable():
                self.terminalFrame.pack_forget()
            if not self.pseudoTerminalFrame.winfo_viewable():
                self.pseudoTerminalFrame.pack(fill=tk.BOTH, expand=True)
            elif self.opened.endswith("|ro"):
                if self.opened in self.pseudoTermFrames:
                    self.pseudoTermFrames[self.opened].pack_forget()
                     

        else:
            if self.pseudoTerminalFrame.winfo_viewable():
                self.pseudoTerminalFrame.pack_forget()
            if not self.terminalFrame.winfo_viewable():
                self.terminalFrame.pack(fill=tk.BOTH, expand=True)
            sessions = self.s.sessions.filter(session_name="pollenisator")
            if sessions:
                session = sessions[0]
                windows = session.windows.filter(window_name=str(iid))
                if not windows:
                    window = session.new_window(attach=True, window_name=str(iid))
                else:
                    window = windows[0]
                    window.select_window()
        self.opened = iid

    def closeDoneTerms(self):
        for iid in self.terminalTv.get_children():
            if iid == "shell":
                continue
            try:
                tags = self.terminalTv.item(str(iid))["tags"]
                if "notified-read" not in tags:
                    continue
            except tk.TclError:
                continue
            if iid.endswith("|ro"):
                if iid in self.pseudoTermFrames:
                    self.pseudoTermFrames[iid].quit()
                self.terminalTv.selection_set("shell")
                self.pseudoTermFrames[iid].destroy()
                del self.pseudoTermFrames[iid]
            else:
                sessions = self.s.sessions.filter(session_name="pollenisator")
                if sessions:
                    session = sessions[0]
                    windows = session.windows.filter(window_name=str(iid))
                    if windows:
                        window = windows[0]
                        window.kill_window()
                    del self.terminalFrames[iid]
            self.terminalTv.delete(str(iid))

    def closeAllTerms(self):
        for iid in self.terminalTv.get_children():
            if iid == "shell":
                continue
            if iid.endswith("|ro"):
                if iid in self.pseudoTermFrames:
                    self.pseudoTermFrames[iid].quit()
                self.terminalTv.selection_set("shell")
                self.pseudoTermFrames[iid].destroy()
                del self.pseudoTermFrames[iid]
            else:
                sessions = self.s.sessions.filter(session_name="pollenisator")
                if sessions:
                    session = sessions[0]
                    windows = session.windows.filter(window_name=str(iid))
                    if windows:
                        window = windows[0]
                        window.kill_window()
                    del self.terminalFrames[iid]
        self.terminalTv.delete(*self.terminalTv.get_children())
        self.terminalTv.insert("", "end", "shell", text="shell", image=TerminalsWidget.getIcon())
        self.terminalTv.selection_set("shell")
            
    def closeTerm(self):
        if self.contextualMenu.selection is None:
            return
        item = self.terminalTv.item(self.contextualMenu.selection)
        iid = str(self.contextualMenu.selection)
        if not iid:
            return
        if iid == "shell":
            return
        
        if iid.endswith("|ro"):
            if iid in self.pseudoTermFrames:
                self.pseudoTermFrames[iid].quit()
            self.terminalTv.selection_set("shell")
            self.pseudoTermFrames[iid].destroy()
            del self.pseudoTermFrames[str(iid)]
        else:
            sessions = self.s.sessions.filter(session_name="pollenisator")
            if sessions:
                session = sessions[0]
                windows = session.windows.filter(window_name=iid)
                if windows:
                    window = windows[0]
                    window.kill_window()
                if str(iid) not in self.terminalFrames:
                    logger.error("TERMINAL not in terminalFrames: "+str(iid)+ " list is "+str(self.terminalFrames))
                else:
                    del self.terminalFrames[str(iid)]
        self.terminalTv.delete(iid)
        


    def create_window(self, iid, title="none", enable_trap=True):
        sessions = self.s.sessions.filter(session_name="pollenisator")
        if sessions:
            session = sessions[0]
            windows = session.windows.filter(window_name=str(iid))
            if iid != "shell":
                session.set_environment("POLLENISATOR_DEFAULT_TARGET", str(iid))
                session.set_environment("TRAP_FOR_POLLEX", str(enable_trap))
            if not windows:
                window = session.new_window(attach=True, window_name=str(iid))
            else:
                window = windows[0]
                #if iid != "shell":
                    #window.panes[0].send_keys("export POLLENISATOR_DEFAULT_TARGET="+str(iid))
                    
            try:
                self.terminalTv.insert("", "end", str(iid), text=title, image=TerminalsWidget.getIcon())
            except tk.TclError:
                pass
            self.terminalTv.selection_set(str(iid)) # trigger treeview select 

    def create_ro_window(self, iid, title, toolController, scanManager):
        tv_iid = iid+"|"+"ro"
        if tv_iid in self.pseudoTermFrames:
            self.pseudoTermFrames[tv_iid].quit()
        self.pseudoTermFrames[tv_iid] = PseudoTermFrame(self.pseudoTerminalFrame, toolController, scanManager)
        self.pseudoTermFrames[tv_iid].pack(fill=tk.BOTH, expand=True)
        try:
            self.terminalTv.insert("", "end", iid=str(tv_iid), text=title, image=TerminalsWidget.getPseudoTerminalIcon())
        except tk.TclError as e:
            print(e)
            pass
        self.terminalTv.selection_set(str(tv_iid)) # trigger treeview select 

        


"""This class pop a tool view form in a subdialog"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.application.dialogs.ChildDialogView import ChildDialogView

class DummyMainApp:
    def __init__(self, settings):
        self.settings = settings

class ChildDialogToolView(ChildDialogView):
    """
    Open a child dialog of a tkinter application to answer a question.
    """
    def __init__(self, parent, title, toolView):
        """
        Open a child dialog of a tkinter application to choose autoscan settings.

        Args:
            parent: the tkinter parent view to use for this window construction.
            toolView : A Tool view object to display
        """
        super().__init__(parent, title)
        self.tool_vw = toolView
        self.tool_vw.appliViewFrame = self.appFrame
        self.tool_vw.openModifyWindow()
        self.completeDialogView(addButtons=False)

   

  


REGISTRY = {}

def register_class(target_class):
    """Register the given class
    Args:
        target_class: type <class>
    """
    REGISTRY[target_class.__name__.lower()] = target_class


class MetaElement(type):
    def __new__(meta, name, bases, class_dict):
        name = name.lower()
        cls = type.__new__(meta, name, bases, class_dict)
        if name not in REGISTRY and name != "element":
            register_class(cls)
        return cls


"""Defines a dialog window to display an error"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import webbrowser
from pollenisatorgui.core.forms.formpanel import FormPanel


def postIssue(_err):
    """Open a tab in a browser to github issues page.
        Args:
            _err: Not used"""
    webbrowser.open_new_tab(
        "https://github.com/AlgoSecure/Pollenisator/issues")


class ChildDialogException:
    """
    Open a child dialog of a tkinter application to present the user to an unhandled exception.
    Can be used to report issue to github.
    """

    def __init__(self, parent, title, err):
        """
        Open a child dialog of a tkinter application to present the user to an unhandled exception.
        Can be used to report issue to github.
        Args:
            parent: the tkinter parent view to use for this window construction.
            title: A title for the new windows
            err: the error that occured causing this window to appear
        """
        self.rvalue = None
        self.parent = parent
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.bind("<Escape>", self.onError)
        self.app.title(title)
        appFrame = CTkFrame(self.app)
        self.form = FormPanel()
        self.err = err
        self.form.addFormLabel(
            "An error occured. Please make an issue with the below stack trace and when it occured.", side=tk.TOP)
        self.form.addFormText("Error", ".+", str(self.err),
                              None, side=tk.TOP, height=500)
        self.form.addFormButton("Report bug", self.onOk, side=tk.RIGHT)
        self.form.addFormButton("Close", self.onError, side=tk.RIGHT)
        self.rvalue = None
        self.form.constructView(appFrame)
        appFrame.pack(ipadx=10, ipady=10)
        self.app.transient(parent)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.focus_force()
            self.app.grab_set()
            self.app.lift()
        except tk.TclError:
            pass

    def onOk(self, _event=None):
        """
        Called when the user clicked the validation button. Set the rvalue attributes to the value selected and close the window.
        
        Args:
            _event: not used but mandatory
        """
        # send the data to the parent
        postIssue(self.err)

    def onError(self, _event=None):
        """
        Close the dialog and set rvalue to None
        Args:
            _event: not used but mandatory
        """
        self.rvalue = False
        self.app.destroy()



"""View for command object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.views.viewelement import ViewElement
from pollenisatorgui.core.components.settings import Settings
import tkinter as tk
import pollenisatorgui.core.components.utils as utils


class CommandView(ViewElement):
    """
    View for command object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory.
    """
    icon = 'command.png'

    def __init__(self, appTw, appViewFrame, mainApp, controller):
        """Constructor
        Args:
            appTw: a PollenisatorTreeview instance to put this view in
            appViewFrame: an view frame to build the forms in.
            mainApp: the Application instance
            controller: a CommandController for this view.
        """
        self.menuContextuel = None
        self.widgetMenuOpen = None
        super().__init__(appTw, appViewFrame, mainApp, controller)

    def _commonWindowForms(self, default={}):
        """Construct form parts identical between Modify and Insert forms
        Args:
            default: a dict of default values for inputs (priority, priority, max_thread). Default to empty respectively "0", "0", "1"
        """
        self.form.addFormHidden("owners", default.get("owners", []))
        panel_bottom = self.form.addFormPanel(grid=True)
        row = 0
        panel_bottom.addFormLabel("Common binary name", row=row)
        panel_bottom.addFormStr("Bin path", r"", default.get("bin_path", ""), column=1, row=row)
        panel_bottom.addFormHelper(
            "The binary name (nmap for Nmap, dirsearch for Diresearch even if it's dir.py for exemple).\nHelps user not to configure this for common name", column=2, row=row)
        row += 1
        
        panel_bottom.addFormLabel("Plugin", row=row)
        panel_bottom.addFormCombo("Plugin", [x["plugin"] for x in APIClient.getInstance().getPlugins()], default.get("plugin", "Default") , column=1, row=row)
        panel_bottom.addFormHelper(
            "The plugin handling this command.", column=2, row=row)
        row += 1
        panel_bottom.addFormLabel("Timeout (in secondes)", row=row)
        panel_bottom.addFormStr("Timeout", r"\d+", default.get("timeout", "300"), width=50, column=1, row=row)
        panel_bottom.addFormHelper(
            "The tool will cancel itself when this duration in second is reached to be run again later.", column=2, row=row)
        row += 1

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing Command
        """
        self.form.clear()
        modelData = self.controller.getData()
        self._initContextualMenu()
        settings = self.mainApp.settings
        settings.reloadSettings()
        self.form.addFormHidden("indb", self.controller.getData()["indb"])
        panel_top = self.form.addFormPanel(grid=True)
        panel_top.addFormLabel("Name", modelData["name"], sticky=tk.NW)
        
        self.panel_text = self.form.addFormPanel()
        self.panel_text.addFormLabel("Command line options", side="top")
        self.panel_text.addFormLabel(
            """Do not include binary name/path\nDo not include Output file option\nRight click to insert variables like ip, scope, port....""", side="right")
        self.panel_text.addFormText("Command line options", r"",
                               modelData["text"], self.menuContextuel, binds={"<Button-3>": self.popup}, side="left", height=100)
        self._commonWindowForms(modelData)
        panel_bottom = self.form.addFormPanel(fill=tk.X, grid=True)
        
        if not self.controller.isMine():
            panel_bottom.addFormButton("Add to my commands", self.controller.addToMyCommands)
        else:
            settings = Settings()
            settings.reloadLocalSettings()
            my_commands = settings.local_settings.get("my_commands", {})
            default = my_commands.get(modelData["name"], modelData["bin_path"])
            panel_bottom.addFormLabel("My binary path")
            panel_bottom.addFormStr("My binary path", r"", default, column=1)
            panel_bottom.addFormHelper("""This is local to your computer. It will not be shared with other users.\nIt is useful to launch commands on your own computer""",column=2)
        
        self.completeModifyWindow(addTags=False)

    def getAdditionalContextualCommands(self):
        return {"Insert command": self.openInsertWindow}

    def openInsertWindow(self):
        """
        Creates a tkinter form using Forms classes. This form aims to insert a new Command
        """
        data = self.controller.getData()
        self._initContextualMenu()
        self.form.addFormHidden("indb", data.get("indb", "pollenisator"))
        panel_top = self.form.addFormPanel(grid=True)
        panel_top.addFormLabel("Name")
        panel_top.addFormStr("Name", r"\S+", data.get("name", ""), None, column=1)
        
        panel_text = self.form.addFormPanel()
        panel_text.addFormLabel("Command line options", side="top")
        panel_text.addFormLabel(
            """Do not include binary name/path\nDo not include Output file option\nRight click to insert variables like ip, scope, port....""", side="right")
        panel_text.addFormText("Command line options",
                               r"", data.get("text", ""), self.menuContextuel, binds={"<Button-3>": self.popup}, side="top", height=100)

        self._commonWindowForms(self.controller.getData())
        self.completeInsertWindow()

    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
        """
        self.appliTw.views[str(self.controller.getDbId())] = {"view": self}
        if parentNode is None:
            parentNode = self.getParentNode()
        tags = self.controller.getTags()
        #if self.controller.isMine():
        #    tags.append("known_command")
        try:
            self.appliTw.insert(parentNode, "end", str(
            self.controller.getDbId()), text=str(self.controller.getModelRepr()), tags=tags, image=self.getClassIcon())
        except tk.TclError:
            pass
        if hasattr(self.appliTw, "hide" ) and self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")
        if "hidden" in tags:
            self.hide("tags")

    def getParentNode(self):
        """
        Return the id of the parent node in treeview.

        Returns:
            return the saved command_node node inside the Appli class.
        """
        return self.appliTw.commands_node

    def _initContextualMenu(self):
        """Initiate contextual menu with variables"""
        self.menuContextuel = utils.craftMenuWithStyle(self.mainApp)
        apiclient = APIClient.getInstance()
        command_variables = apiclient.getCommandVariables()
        for variable_name in command_variables:
           self.menuContextuel.add_command(label=variable_name, command=lambda variable_name=variable_name: self.addVariable(variable_name))
        # self.menuContextuel.add_command(
        #     label="Wave id", command=self.addWaveVariable)
        # self.menuContextuel.add_command(
        #     label="Network address without slash nor dots", command=self.addIpReseauDirVariable)
        # self.menuContextuel.add_command(
        #     label="Network address", command=self.addIpReseauVariable)
        # self.menuContextuel.add_command(
        #     label="Parent domain", command=self.addParentDomainVariable)
        # self.menuContextuel.add_command(label="Ip", command=self.addIpVariable)
        # self.menuContextuel.add_command(
        #     label="Ip without dots", command=self.addIpDirVariable)
        # self.menuContextuel.add_command(
        #     label="Port", command=self.addPortVariable)

    def addVariable(self, variable):
        """
        insert the variable inside the a tkinter widget stored in appli widgetMenuOpen attribute.

        Args:
            variable: the variable to insert
        """
        self.widgetMenuOpen.insert(tk.INSERT, "|"+variable+"|")

    def popup(self, event):
        """
        Fill the self.widgetMenuOpen and reraise the event in the editing window contextual menu

        Args:
            event: a ttk Treeview event autofilled. Contains information on what treeview node was clicked.
        """
        print("command view popup")
        self.widgetMenuOpen = event.widget
        self.menuContextuel.tk_popup(event.x_root, event.y_root)
        self.menuContextuel.focus_set()
        self.menuContextuel.bind('<FocusOut>', self.popupFocusOut)

    def popupFocusOut(self, _event=None):
        """
        Called when the contextual menu is unfocused
        Args:
            _event: a ttk event autofilled. not used but mandatory.
        """
        self.menuContextuel.unpost()

    def addWaveVariable(self):
        """
        insert the wave variable inside the a tkinter widget stored in appli widgetMenuOpen attribute.
        """
        self.widgetMenuOpen.insert(tk.INSERT, "|wave|")

    def addIpReseauDirVariable(self):
        """
        insert the scope_dir variable inside the a tkinter widget stored in appli widgetMenuOpen attribute.
        """
        self.widgetMenuOpen.insert(tk.INSERT, "|scope_dir|")

    def addIpReseauVariable(self):
        """
        insert the scope variable inside the a tkinter widget stored in appli widgetMenuOpen attribute.
        """
        self.widgetMenuOpen.insert(tk.INSERT, "|scope|")

    def addParentDomainVariable(self):
        """
        insert the scope variable inside the a tkinter widget stored in appli widgetMenuOpen attribute.
        """
        self.widgetMenuOpen.insert(tk.INSERT, "|parent_domain|")

    def addIpVariable(self):
        """
        insert the ip variable inside the a tkinter widget stored in appli widgetMenuOpen attribute.
        """
        self.widgetMenuOpen.insert(tk.INSERT, "|ip|")

    def addIpDirVariable(self):
        """
        insert the ip_dir variable inside the a tkinter widget stored in appli widgetMenuOpen attribute.
        """
        self.widgetMenuOpen.insert(tk.INSERT, "|ip_dir|")

    def addPortVariable(self):
        """
        insert the port variable inside the a tkinter widget stored in appli widgetMenuOpen attribute.
        """
        self.widgetMenuOpen.insert(tk.INSERT, "|port|")


    def updateReceived(self, obj=None, old_obj=None):
        """Called when a command update is received by notification.
        Update the command treeview item (resulting in icon reloading)
        """
        # if self.controller.isMine():
        #     try:
        #         self.appliTw.move(str(self.controller.model.getId()), self.appliTw.my_commands_node, "end")
        #     except tk.TclError:
        #         print("WARNING: Update received for a non existing command "+str(self.controller.getModelRepr()))
        # else:
        #     try:
        #         self.appliTw.move(str(self.controller.model.getId()), self.appliTw.commands_node, "end")
        #     except tk.TclError:
        #         print("WARNING: Update received for a non existing command "+str(self.controller.getModelRepr()))
        
        super().updateReceived()

    def key(self):
        """Returns a key for sorting this node
        Returns:
            tuple, key to sort
        """
        return tuple([ord(c) for c in str(self.controller.getModelRepr()).lower()])



"""Help the user to create a new pentest database.
"""
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from datetime import datetime
from pollenisatorgui.core.components import utils
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.views.viewelement import ViewElement
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.components.apiclient import APIClient

class ChildDialogNewPentest(CTkToplevel):
    """
    Open a child dialog of a tkinter application to ask details about
    a new pentest database to create.
    """

    def __init__(self, parent, default):
        """
        Open a child dialog of a tkinter application to ask details about
        the new pentest.

        Args:
            parent: the tkinter parent view to use for this window construction.
        """
        super().__init__(parent)
        self.attributes("-type", "dialog")
        self.parent = parent
        self.title("Create a pentest")
        self.resizable(True, True)
        self.bind("<Escape>", self.onError)
        monitor = utils.get_screen_where_widget(parent)
        self.geometry(f"{800}x{900}+{monitor.x+(monitor.width//2)-(800//2)}+{monitor.y+(monitor.height//2)-(900//2)}")
        
        self.rvalue = None
        self.parent = parent
        mainFrame = CTkFrame(self)
        self.form = FormPanel()
        form1 = self.form.addFormPanel(grid=True, side=tk.TOP, fill=tk.X)
        form1.addFormLabel("Pentest name")
        form1.addFormStr("Pentest name", r".+", default=default.get("name", ""), column=1)
        form1.addFormLabel("Mission name",row=1)
        form1.addFormStr("Mission name", r".+", default=default.get("mission_name", ""), row=1, column=1)
        form1.addFormLabel("Client's name", row=2)
        form1.addFormStr("Client's name", r".+", default=default.get("client_name", ""), row=2, column=1)
        form1.addFormLabel("Report language", row=3)
        langs = APIClient.getInstance().getLangList()
        form1.addFormCombo("lang", langs, default=default.get("lang", "en"), row=3, column=1)
        types = list(Settings.getPentestTypes().keys())
        if types:
            form1.addFormLabel("Pentest type", row=4)
            form1.addFormCombo(
                "Pentest type", types, default=default.get("type", types[0]), row=4, column=1)
        form1.addFormLabel("Starting", row=5)
        form1.addFormDate("startd", self, default.get("start", datetime.strftime(datetime.now(), "%d/%m/%Y %H:%M:%S")), "%d/%m/%Y %H:%M:%S", row=5, column=1)
        form1.addFormLabel("Ending", row=6)
        form1.addFormDate("endd", self, default.get("end", "31/12/2099 00:00:00"),
                          "%d/%m/%Y %H:%M:%S", row=6, column=1)
        form2 = self.form.addFormPanel(grid=True, side=tk.TOP, fill=tk.X, pady=5)
        form2.addFormLabel("Scope", pady=5)
        form2.addFormText(
            "Scope", "", default.get("scope", ""),  height=100, column=1, sticky=tk.W+tk.E, pady=5)
        form2.addFormHelper(
            "You can declare network ip as IP/MASKSIZE, ips or domains", column=2, sticky=tk.W)
        form3 = self.form.addFormPanel(side=tk.TOP, fill=tk.X, pady=5)
        form3.addFormSearchBar("Pentester search", self.searchCallback, self.form, side=tk.TOP)
        form3.addFormLabel("Pentesters added", side=tk.LEFT)
        form3.addFormTreevw(
            "Additional pentesters names", ("Additional pentesters names", "added"), default.get("pentesters", ["", ""]), add_empty_row=True, width=200, pady=5, fill=tk.X, side=tk.TOP, anchor=tk.CENTER)
        form4 = self.form.addFormPanel(side=tk.TOP, fill=tk.X, pady=5)
        default_settings = []
        for key, val in default.get("settings", {}).items():
            if val == 1:
                default_settings.append(key)
        form4.addFormChecklist("Settings", ["Add domains whose IP are in scope",
                                            "Add domains who have a parent domain in scope", "Add all domains found"], default_settings, side=tk.TOP, fill=tk.X, pady=5)
        form4.addFormButton("Create", self.onOk, side=tk.BOTTOM)
        self.form.constructView(mainFrame)
        form1.setFocusOn("Database name")
        mainFrame.pack(fill=tk.BOTH, ipadx=10, ipady=10)
        try:
            self.wait_visibility()
            self.transient(parent)
            self.focus_force()
            self.grab_set()
            self.lift()
        except tk.TclError:
            pass
        self.update()

    def searchCallback(self, searchreq):
        apiclient = APIClient.getInstance()
        users = apiclient.searchUsers(searchreq)
        if users is None:
            return [], "Invalid response from API"
        ret = [{"TITLE":user, "additional pentesters names":{"text":user}} for user in users]
        return ret, ""
    
    def onError(self, _event=None):
        self.rvalue = None
        self.destroy()

    def onOk(self, _event=None):
        """
        Called when the user clicked the validation button. Set the rvalue attributes to the value selected and close the window.
        
        Args:
            _event: not used but mandatory
        """
        # send the data to the parent
        res, msg = self.form.checkForm()
        if res:
            form_values = self.form.getValue()
            form_values_as_dicts = ViewElement.list_tuple_to_dict(form_values)
            self.rvalue = {"name": form_values_as_dicts["Pentest name"],
                           "mission_name": form_values_as_dicts["Mission name"],
                           "client_name": form_values_as_dicts["Client's name"],
                           "lang": form_values_as_dicts["lang"],
                           "type": form_values_as_dicts.get("Pentest type", ""),
                           "start": form_values_as_dicts["startd"],
                           "end": form_values_as_dicts["endd"],
                           "settings": form_values_as_dicts["Settings"],
                           "scope": form_values_as_dicts["Scope"],
                           "pentesters": "\n".join([x[0] for x in form_values_as_dicts["Additional pentesters names"] if x and x[0] != ""])}
            self.destroy()
        else:
            tk.messagebox.showwarning(
                "Form not validated", msg, parent=self)



"""CheckItem Model."""
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.core.components.apiclient import APIClient
from bson import ObjectId

class CheckItem(Element):
    """Represents a CheckItem object of a cheatsheet.

    Attributes:
        coll_name: collection name in pollenisator or pentest database
    """

    coll_name = "checkitems"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched command is optimal.
                        possible keys with default values are : _id (None), parent (None), title, pentest_types=None, description="", category="", commands=None, defect_tags=None, script="",infos=None
        """
        if valuesFromDb is None:
            valuesFromDb = dict()
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None), valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("title"), valuesFromDb.get("pentest_types", []), valuesFromDb.get("lvl", "port:onServiceUpdate"),
            valuesFromDb.get("ports", ""), valuesFromDb.get("priority", 0), valuesFromDb.get("max_thread", 1), valuesFromDb.get("description", ""), valuesFromDb.get("category", ""),\
            valuesFromDb.get("check_type", "manual"), valuesFromDb.get("step", 0), valuesFromDb.get("parent", None),
            valuesFromDb.get("commands", []), valuesFromDb.get("defect_tags", []), valuesFromDb.get("script", ""),  valuesFromDb.get("infos", {}))
        
    def initialize(self, title, pentest_types=None, lvl="port:onServiceUpdate", ports="", priority=0, max_thread=1, description="", category="", check_type="manual", step=0, parent=None, commands=None, defect_tags=None, script="", infos=None):
        self.title = title
        self.description = description
        self.category = category
        self.check_type = check_type
        self.step = step
        self.parent = parent
        self.lvl = lvl
        self.ports = ports
        self.max_thread = int(max_thread)
        self.priority = int(priority)
        self.commands = [] if commands is None else commands
        self.defect_tags = [] if defect_tags is None else defect_tags
        self.script = script
        self.pentest_types = [] if pentest_types is None else pentest_types
        self.defect_tags = [] if defect_tags is None else defect_tags
        self.infos = {} if infos is None else infos
        return self

    def delete(self):
        """
        Delete the command represented by this model in database.
        Also delete it from every waves's wave_commands
        Also delete every tools refering to this command.
        """
        ret = self._id
        apiclient = APIClient.getInstance()
        apiclient.deleteCheckItem(ret)
        

    def addInDb(self):
        """Add this command to pollenisator database
        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        apiclient = APIClient.getInstance()
        res, id = apiclient.insertCheckItem(self.getData())
        if not res:
            return False, id
        self._id = id
        return True, id
        

    def update(self, pipeline_set=None):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        if pipeline_set is None:
            apiclient.updateCheckItem(self._id, self.getData())
        else:
            apiclient.updateCheckItem(self._id, pipeline_set )
        


    @classmethod
    def fetchObject(cls,  pipeline):
        """Fetch one CheckItem from database and return the CheckItem object 
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a CheckItem or None if nothing matches the pipeline.
        """
        apiclient = APIClient.getInstance()
        res = apiclient.findCheckItem(pipeline, many=False)
        d = res
        if d is None:
            return None
        return CheckItem(d)

    @classmethod
    def fetchObjects(cls, pipeline):
        """Fetch many cheatsheet from database and return a Cursor to iterate over cheatsheet objects
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a cursor to iterate on cheatsheet objects
        """
        apiclient = APIClient.getInstance()
        ds = apiclient.findCheckItem(pipeline)
        if ds is None:
            return None
        for d in ds:
            yield CheckItem(d)
    
    def getChecks(self):
        """Return check instances that implements this check item
        Returns:
            list of checkInstance objects
        """
        datamanager = DataManager.getInstance()
        return datamanager.find("checkinstance", {"check_iid": str(self._id)}, multi=True)
    
    def getData(self):
        return {"_id": self._id,  "title":self.title, "pentest_types":self.pentest_types, "lvl":self.lvl, "ports":self.ports, 
                "priority": int(self.priority), "max_thread": int(self.max_thread),
                "description": self.description, "category":self.category,
                "check_type":self.check_type, "step":self.step, "parent":self.parent,
                "commands":self.commands, "defect_tags":self.defect_tags, "script":self.script, "infos":self.infos}


    def getDbKey(self):
        """Return a dict from model to use as unique composed key.
        Returns:
            A dict (1 key :"name")
        """
        return {"title": self.title}

    def isAuto(self):
        """Return True if this command is an auto command"""
        return "auto" in self.check_type

    def __str__(self):
        return self.title

    @classmethod
    def fetchPentestObjects(cls):
        return cls.fetchObjects({})


"""Port Model"""

from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.core.models.tool import Tool
from pollenisatorgui.core.models.defect import Defect
from pollenisatorgui.core.components.apiclient import APIClient
from bson.objectid import ObjectId
import webbrowser

class Port(Element):
    """
    Represents an Port object that defines an Port that will be targeted by port level tools.

    Attributes:
        coll_name: collection name in pollenisator database
    """
    coll_name = "ports"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched interval is optimal.
                        possible keys with default values are : _id (None), parent (None), tags([]), infos({}),
                        ip(""), port(""), proto("tcp"), service(""), product(""), notes("")
        """
        if valuesFromDb is None:
            valuesFromDb = {}
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None),  valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("ip", ""), valuesFromDb.get("port", ""),
                        valuesFromDb.get("proto", "tcp"), valuesFromDb.get(
                            "service", ""), valuesFromDb.get("product", ""),
                        valuesFromDb.get("notes", ""), valuesFromDb.get("infos", {}))

    def initialize(self, ip, port="", proto="tcp", service="", product="", notes="",infos=None):
        """Set values of port
        Args:
            ip: the parent host (ip or domain) where this port is open
            port: a port number as string. Default ""
            proto: a protocol to reach this port ("tcp" by default, send "udp" if udp port.) Default "tcp"
            service: the service running behind this port. Can be "unknown". Default ""
            notes: notes took by a pentester regarding this port. Default ""
            infos: a dictionnary of additional info. Default is None (empty dict)
        Returns:
            this object
        """
        self.ip = ip
        self.port = port
        self.proto = proto
        self.service = service
        self.product = product
        self.notes = notes
        self.infos = infos if infos is not None else {}
        return self

    def delete(self):
        """
        Deletes the Port represented by this model in database.
        Also deletes the tools associated with this port
        Also deletes the defects associated with this port
        """
        apiclient = APIClient.getInstance()
        
        apiclient.delete("ports", ObjectId(self._id))

    def update(self, pipeline_set=None):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        # Update variable instance. (this avoid to refetch the whole command in database)
        if pipeline_set is None:
            apiclient.update("ports", ObjectId(self._id), {"service": self.service, "product":self.product, "notes": self.notes,  "infos": self.infos})
        else:
            apiclient.update("ports", ObjectId(self._id),  pipeline_set)

    def addInDb(self):
        """
        Add this Port in database.

        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        base = self.getDbKey()
        apiclient = APIClient.getInstance()
        # Inserting port
        base["service"] = self.service
        base["product"] = self.product
        base["notes"] = self.notes
        base["infos"] = self.infos
        res, iid = apiclient.insert("ports", base)
        self._id = iid
        
        return res, iid

    def addCustomTool(self, command_iid):
        """
        Add the appropriate tools (level check and wave's commands check) for this port.

        Args:
            command_iid: The command that we want to create all the tools for.
        """
        apiclient = APIClient.getInstance()
        return apiclient.addCustomTool(self._id, command_iid)

    def _getParentId(self):
        """
        Return the mongo ObjectId _id of the first parent of this object. For a port it is the ip.

        Returns:
            Returns the parent ip's ObjectId _id".
        """
        datamanager = DataManager.getInstance()
        obj = datamanager.find("ips", {"ip":self.ip}, False)
        if obj is None:
            return None
        return obj.getId()

    def __str__(self):
        """
        Get a string representation of a port.

        Returns:
            Returns the string protocole/port number.
        """
        return (self.proto+"/" if self.proto == "udp" else "")+str(self.port)+" "+str(self.service)

    def getDetailedString(self):
        """Returns a detailed string describing this port.
        Returns:
            string : ip:proto/port
        """
        return str(self.ip)+":"+str(self)

    def getURL(self):
        ssl = self.infos.get("SSL", None) == "True" or ("https" in self.service or "ssl" in self.service)
        url = "https://" if ssl else "http://"
        url += self.ip+":"+self.port+"/"
        return url

    def getDefects(self):
        """Return port assigned defects as a list of mongo fetched defects dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        datamanager = DataManager.getInstance()
        return datamanager.find("defects", {"ip": self.ip, "port": self.port, "proto": self.proto})

    def getDbKey(self):
        """Return a dict from model to use as unique composed key.
        Returns:
            A dict (3 keys :"ip", "port", "proto")
        """
        return {"ip": self.ip, "port": self.port, "proto": self.proto}

    def getData(self):
        """Return port attributes as a dictionnary matching Mongo stored ports
        Returns:
            dict with keys ip, port, proto, service, product, notes, _id, tags and infos
        """
        return {"ip": self.ip, "port": self.port, "proto": self.proto,
                "service": self.service, "product": self.product, "notes": self.notes, "_id": self.getId(), "infos": self.infos}

    def openInBrowser(self, _event=None):
        """Callback for action open in browser
        Args:
            _event: nut used but mandatory
        """
        if self.service == "http":
            webbrowser.open_new_tab(
                "http://"+self.ip+":"+self.port)
        else:
            webbrowser.open_new_tab(
                "https://"+self.ip+":"+self.port)


import os
import uuid
import socket
from pollenisatorgui.core.components.scanworker import ScanWorker
from pollenisatorgui.core.components.settings import Settings

def pollworker():
    """Starts a worker that receives scan orders from the server and upload results
    """
    settings = Settings()
    settings.reloadLocalSettings()
    sm = ScanWorker(settings)
    myname = os.getenv('POLLENISATOR_WORKER_NAME', str(uuid.uuid4())+"@"+socket.gethostname())
    sm.connect(myname)
    sm.wait()    



"""Describe tkinter Text with default common args and an attached scrollbar"""

import tkinter as tk
from customtkinter import *
import pyperclip
from pollenisatorgui.core.forms.form import Form
import pollenisatorgui.core.components.utils as utils

class FormText(Form):
    """
    Form field representing a multi-lined input.
    Default setted values:
        width=20, height=20
        if pack : padx = pady = 5, side = left
        if grid: row = column = 0 sticky = "East"
    """

    def __init__(self, name, validation="", default="", contextualMenu=None, **kwargs):
        """
        Constructor for a form text

        Args:
            name: the entry name (id).
            validation: a regex used to check the input
                             in the checkForm function. default is "" or a callback function
            default: a default value for the Entry, default is ""
            contextualMenu: (Opt.) a contextualMenu to open when right clicked. default is None
            kwargs: same keyword args as you would give to ttk.Text
        """
        super().__init__(name)
        self.validation = validation
        self.default = default
        self.contextualMenu = contextualMenu
        self.kwargs = kwargs
        self.binds = self.getKw("binds",{})
        self.text = None
        self.widgetMenuOpen = None

    def _initContextualMenu(self, parent):
        """Initialize the contextual menu for paperclip.
        Args:
            parent: the tkinter parent widget for the contextual menu
        """
        # FIXME Add to given menu instead of Overriding given contextual menu
        if self.contextualMenu is None:
            self.contextualMenu = utils.craftMenuWithStyle(parent)
        parent.bind("<Button-3>",self.binds.get("<Button-3>", self.popup))
        self.contextualMenu.add_command(label="Copy", command=self.copy)
        self.contextualMenu.add_command(label="Cut", command=self.cut)
        self.contextualMenu.add_command(label="Paste", command=self.paste)
        self.contextualMenu.add_command(label="Close", command=self.close)

    def close(self):
        """Option of the contextual menu : Close the contextual menu by doing nothing
        """
        pass

    def copy(self):
        """Option of the contextual menu : Copy entry text to clipboard
        """
        ranges = self.text.tag_ranges(tk.SEL)
        if ranges:
            pyperclip.copy(self.text.get(*ranges))

    def cut(self):
        """Option of the contextual menu : Cut entry text to clipboard
        """
        ranges = self.text.tag_ranges(tk.SEL)
        if ranges:
            pyperclip.copy(self.text.get(*ranges))
            self.text.delete(*ranges)

    def paste(self, event=None):
        """Option of the contextual menu : Paste clipboard content to entry
        """
        if event is None:
            buff = pyperclip.paste()
        else:
            buff = event.widget.clipboard_get()
        if buff:
            #delete selected text
            ranges = self.text.tag_ranges(tk.SEL)
            if ranges:
                self.text.delete(*ranges)
            #insert clipboard content
            insertIndex = self.text.index(tk.INSERT)
            self.text.insert(insertIndex, self.sanitize(buff))
        return "break" # break the event to prevent the default paste
    
    def sanitize(self, string):
        """remove unwanted things in text"""
        return string.replace("\r","")

    def constructView(self, parent):
        """
        Create the text view inside the parent view given

        Args:
            parent: parent FormPanel.
        """
        state = self.getKw("state", "normal")
        self.text = CTkTextbox(
            parent.panel, height=self.getKw("height", 200), width=self.getKw("width", 500),border_width=1, wrap="word")
        self._initContextualMenu(self.text)
        self.text.bind('<Control-a>', self.selectAll)
        self.text.bind("<<Paste>>", self.paste)
        try:
            self.text.insert(tk.INSERT, self.sanitize(self.default))
        except tk.TclError as e:
            self.text.insert(tk.INSERT, "Error :\n"+str(e))
        if state == "disabled":
            self.text.configure(state="disabled")
        if parent.gridLayout:
            self.text.grid(row=self.getKw("row", 0), column=self.getKw(
                "column", 0), sticky=self.getKw("sticky", tk.EW), **self.kwargs)
        else:
            self.text.pack(side=self.getKw("side", "left"), padx=self.getKw(
                "padx", 10), pady=self.getKw("pady", 5), expand=self.getKw("expand", True), fill=self.getKw("fill", "both"), **self.kwargs)

    def selectAll(self, _event=None):
        """Callback to select all the text in the date Entry.
        Args:
            _event: mandatory but not used
        Returns:
            Returns the string "break" to prevent the event to be treated by the Entry, thus inserting unwanted value.
        """
        self.text.tag_add('sel', '1.0', 'end')
        return "break"

    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return the entry value as string.
        """
        return self.text.get('1.0', 'end-1c')

    def setValue(self, newval):
        """
        Set the text value.
        Args:
            newval: the new value to be set inside the text
        """
        state = self.text._textbox.cget("state")
        self.text.configure(state="normal")
        self.text.delete("1.0", "end")
        self.text.insert("1.0", self.sanitize(newval))
        self.text.configure(state=state)

    def checkForm(self):
        """
        Check if this form is correctly filled.
        Check with the validation given in constructor.

        Returns:
            {
                "correct": True if the form is correctly filled, False otherwise.
                "msg": A message indicating what is not correctly filled.
            }
        """
        if isinstance(self.validation, str):
            import re
            if re.match(self.validation, self.getValue(), re.MULTILINE) is None:
                return False, self.name+" value is incorrect."
            return True, ""
        elif callable(self.validation):
            return self.validation(self.getValue()), self.name+" value is incorrect."



    def popup(self, event):
        """
        Fill the self.widgetMenuOpen and reraise the event in the editing window contextual menu

        Args:
            event: a ttk Treeview event autofilled.
            Contains information on what treeview node was clicked.
        """
        self.widgetMenuOpen = event.widget
        self.contextualMenu.tk_popup(event.x_root, event.y_root)
        self.contextualMenu.focus_set()
        self.contextualMenu.bind('<FocusOut>', self.popupFocusOut)

    def popupFocusOut(self, _event=None):
        """Callback for focus out event. Destroy contextual menu
        Args:
            _event: not used but mandatory
        """
        self.contextualMenu.unpost()

    def setFocus(self):
        """Set the focus to the ttk entry widget.
        """
        self.text.focus_set()



"""This class pop a dialog to view scan history"""

from datetime import datetime
import tkinter as tk
from customtkinter import *
from pollenisatorgui.core.application.scrollableframexplateform import ScrollableFrameXPlateform
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.application.scrollabletreeview import ScrollableTreeview
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.models.checkinstance import CheckInstance
from pollenisatorgui.core.models.tool import Tool
from pollenisatorgui.core.controllers.toolcontroller import ToolController 
from pollenisatorgui.core.views.toolview import ToolView
from pollenisatorgui.core.application.dialogs.ChildDialogToolView import ChildDialogToolView
from bson.objectid import ObjectId

class ChildDialogScanHistory:
    """
    Open a child dialog of a tkinter application
    """
    def __init__(self, parent):
        """
        Open a child dialog of a tkinter application to choose autoscan checks.

        Args:
            parent: the tkinter parent view to use for this window construction.
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        #self.app.geometry("800x650")
        self.app.title("Scan history")
        self.app.resizable(True, True)
        self.app.bind("<Escape>", self.cancel)
        self.rvalue = None
        appFrame = CTkFrame(self.app)
        self.parent = parent
        self.initUI(appFrame)
        frame_buttons = CTkFrame(self.app)
        ok_button = CTkButton(frame_buttons, text="OK")
        ok_button.pack(side=tk.RIGHT, padx=5)
        ok_button.bind('<Button-1>', self.okCallback)
        cancel_button = CTkButton(frame_buttons, text="Cancel", 
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        cancel_button.pack(side=tk.RIGHT, padx=5)
        cancel_button.bind('<Button-1>', self.cancel)
        frame_buttons.pack(side=tk.BOTTOM, anchor=tk.SE , padx=5 , pady=5)
        appFrame.pack(fill=tk.BOTH, pady=10, padx=10, expand="yes")

        self.app.transient(parent)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.grab_set()
            self.app.focus_force()
            self.app.lift()
        except tk.TclError:
            pass

    def cancel(self, _event=None):
        """called when canceling the window.
        Close the window and set rvalue to False
        Args:
            _event: Not used but mandatory"""
        self.rvalue = None
        self.app.destroy()

    def okCallback(self, _event=None):
        """called when pressing the validating button
        Close the window if the form is valid.
        Set rvalue to True and perform the defect update/insert if validated.
        Args:
            _event: Not used but mandatory"""
        
        self.rvalue = []
        self.app.destroy()

    def OnDoubleClick(self, event):
        """
        Callback for double click on treeview.
        Opens a window to update the double clicked tool view.
        Args:
            event: automatically created with the event catch. stores data about line in treeview that was double clicked.
        """
        item = self.histoScanTv.identify("item", event.x, event.y)
        if item is None or item == '':
            return
        datamanager = DataManager.getInstance()
        tool = datamanager.get("tool", str(item))
        if tool is None:
            return
        
        dialog = ChildDialogToolView(self.app, "Tool view", ToolView(None, None, None, ToolController(tool)))
            
        
    def filter(self, event=None):
        category = self.str_filter_category.getValue()
        name = self.str_filter_name.getValue()
        startd = self.date_filter_startd.getValue()
        endd = self.date_filter_endd.getValue()
        self.histoScanTv.filter(category, name, (ScrollableTreeview.date_compare, startd, endd))

    def initUI(self, frame):
        form = FormPanel(side=tk.TOP, fill=tk.X, pady=5, grid=True)
        form.addFormLabel("Category")
        self.str_filter_category = form.addFormStr("Category", placeholder_text="category", column=1, binds={"<Key-Return>":  self.filter})
        form.addFormLabel("Name", column=2)
        self.str_filter_name = form.addFormStr("Name", placeholder_text="name", column=3, binds={"<Key-Return>":  self.filter})
        form.addFormLabel("Start date", column=4)
        start = datetime.today()
        start = start.replace(hour=0, minute=0, second=0, microsecond=0)
        self.date_filter_startd = form.addFormDate("Start Date", self.app, default=utils.dateToString(start), column=5, binds={"<Key-Return>":  self.filter})
        form.addFormLabel("End date", column=6)
        self.date_filter_endd = form.addFormDate("End Date", self.app, default=utils.dateToString(datetime.now()), column=7, binds={"<Key-Return>":  self.filter})
        form.constructView(frame)

        self.histoScanTv = ScrollableTreeview(frame, ('History category', 'Name', 'Ended at'), height=25, sort_keys=(None, None, utils.stringToDate))
        self.histoScanTv.pack(fill=tk.BOTH, expand="yes", padx=10, pady=10)
        self.histoScanTv.bind("<Double-Button-1>", self.OnDoubleClick)
        self.refreshUI()

    
    def refreshUI(self):
        done_scans = list(Tool.fetchObjects({"status":"done"}))
        checks = CheckInstance.fetchObjects([ObjectId(done_scan.check_iid) for done_scan in done_scans if done_scan.check_iid != ""])
        self.mapping = {}
        for check in checks:
            self.mapping[str(check._id)] = check
        self.histoScanTv.reset()
        for done_scan in done_scans:
            check = self.mapping.get(str(done_scan.check_iid), None)
            group_name = "" if check is None else check.check_m.title
            try:
                self.histoScanTv.insert('',"end", str(done_scan.getId()), text=group_name, values=(done_scan.name, done_scan.datef))
            except tk.TclError as e:
                print(e)
        



"""Controller for interval object. Mostly handles conversion between mongo data and python objects"""

from pollenisatorgui.core.controllers.controllerelement import ControllerElement


class IntervalController(ControllerElement):
    """Inherits ControllerElement
    Controller for interval object. Mostly handles conversion between mongo data and python objects"""

    def doUpdate(self, values):
        """
        Update the Interval represented by this model in database with the given values.

        Args:
            values: A dictionary crafted by IntervalView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated interval document.
        """
        self.model.dated = values.get("Start date", self.model.dated)
        self.model.datef = values.get("End date", self.model.datef)
        self.model.update()

    def doInsert(self, values):
        """
        Insert the Interval represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by IntervalView containg all form fields values needed.

        Returns:
            {
                '_id': The mongo ObjectId _id of the inserted command document.
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        # Get form values
        dated = values["Start date"]
        datef = values["End date"]
        # Insert in database
        self.model.initialize(values["waveName"], dated, datef)
        ret, _ = self.model.addInDb()
        return ret, 0  # 0 errors

    
    def getType(self):
        """Return a string describing the type of object
        Returns:
            "interval" """
        return "interval"


"""Help the user to create a setup an autoscan.
"""
import tkinter as tk
from tkinter import ttk
from customtkinter import *
from pollenisatorgui.core.application.checkboxscrollabletreeview import CheckboxScrollableTreeview
from pollenisatorgui.core.application.scrollabletreeview import ScrollableTreeview
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.settings import Settings
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.models.checkitem import CheckItem
from pollenisatorgui.core.models.command import Command
from pollenisatorgui.core.components.logger_config import logger


class ChildDialogAutoScanParams(CTkToplevel):
    """
    Open a child dialog of a tkinter application to ask details about
    a new pentest database to create.
    """

    def __init__(self, parent):
        """
        Open a child dialog of a tkinter application to choose autoscan checks.

        Args:
            parent: the tkinter parent view to use for this window construction.
        """
        super().__init__(parent)
        self.attributes("-type", "dialog")
        self.title("AutoScan configuration")
        self.resizable(True, True)
        self.bind("<Escape>", self.cancel)
        self.rvalue = None
        appFrame = CTkFrame(self)
        self.parent = parent
        self.initUI(appFrame)
        frame_buttons = CTkFrame(self)
        ok_button = CTkButton(frame_buttons, text="OK")
        ok_button.pack(side=tk.RIGHT, padx=5)
        ok_button.bind('<Button-1>', self.okCallback)
        cancel_button = CTkButton(frame_buttons, text="Cancel", 
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        cancel_button.pack(side=tk.RIGHT, padx=5)
        cancel_button.bind('<Button-1>', self.cancel)
        frame_buttons.pack(side=tk.BOTTOM, anchor=tk.SE , padx=5 , pady=5)
        appFrame.pack(fill=tk.BOTH, ipady=10, ipadx=10, expand="yes")

        self.transient(parent)
        try:
            self.wait_visibility()
            self.transient(parent)
            self.grab_set()
            self.focus_force()
            self.lift()
        except tk.TclError:
            pass

    def cancel(self, _event=None):
        """called when canceling the window.
        Close the window and set rvalue to False
        Args:
            _event: Not used but mandatory"""
        self.rvalue = None
        self.destroy()

    def okCallback(self, _event=None):
        """called when pressing the validating button
        Close the window if the form is valid.
        Set rvalue to True and perform the defect update/insert if validated.
        Args:
            _event: Not used but mandatory"""
        
        self.rvalue = {}
        self.rvalue["commands"] = self.autoScanTv.get_checked_children()
        self.rvalue["autoqueue"] = self.switch_auto_queue.get()
        self.destroy()


    def filter(self, event=None):
        category = self.str_filter_category.getValue()
        title = self.str_filter_category_title.getValue()
        prio = self.str_filter_priority.getValue()
        command = self.str_filter_command.getValue()
        self.autoScanTv.filter(command, category, title, (lambda prio, toCompare: int(toCompare)<=int(prio) if prio != "" else True, prio), check_case=False)

    def initUI(self, frame):
        form = FormPanel(side=tk.TOP, fill=tk.X, pady=5, grid=True)
        column = 0
        form.addFormLabel("Command name", column=(column := column + 1))
        self.str_filter_command = form.addFormStr("Command", placeholder_text="Nmap", column=(column := column + 1), binds={"<Key-Return>":  self.filter})
        form.addFormLabel("Category", column=(column := column + 1))
        self.str_filter_category = form.addFormStr("Category", placeholder_text="recon", column=(column := column + 1), binds={"<Key-Return>":  self.filter})
        form.addFormLabel("Category title", column=(column := column + 1))
        self.str_filter_category_title = form.addFormStr("CategoryTitle", placeholder_text="title", column=(column := column + 1), binds={"<Key-Return>":  self.filter})
        form.addFormLabel("Priority", column=(column := column + 1))
        self.str_filter_priority = form.addFormStr("Priority", placeholder_text="0", column=(column := column + 1), binds={"<Key-Return>":  self.filter})
        form.constructView(frame)
        self.check_box_all = CTkCheckBox(frame, text="All", command=self.check_all)
        self.check_box_all.pack(padx=10,pady=10,side=tk.TOP)
        tvFrame = CTkFrame(frame)
        self.autoScanTv = CheckboxScrollableTreeview(tvFrame, ("Command", 'Category', 'Category title', 'Priority'), maxPerPage=25, height=25, sort_keys=(None, None, None, int), fill=tk.X, autoresize=False)
        self.autoScanTv.pack(fill=tk.BOTH, expand="yes", padx=10, side=tk.BOTTOM, anchor=tk.CENTER)
        tvFrame.pack(fill=tk.BOTH, expand="yes", padx=10, pady=10, side=tk.BOTTOM)
        self.switch_auto_queue = CTkSwitch(frame, text="Continously queue selected command(s)")
        self.switch_auto_queue.pack(padx=10,pady=15,side=tk.BOTTOM)
        self.refreshUI()

    def check_all(self):
        if self.check_box_all.get():
            self.autoScanTv.checkAll()
        else:
            self.autoScanTv.uncheckAll()
    
    def refreshUI(self):
        settings = Settings()
        settings._reloadDbSettings()
        apiclient = APIClient.getInstance()
        checks = list(CheckItem.fetchObjects({"pentest_types":settings.getPentestType(), "check_type":"auto_commands"}))
        commands_iids = set()
        lkuptable = {}
        for check in checks:
            commands_iids = commands_iids.union(set(check.commands))
            for command in check.commands:
                lkuptable[str(command)] = check
        commands = Command.fetchObjects({"original_iid":{"$in":list(commands_iids)}}, targetdb=apiclient.getCurrentPentest())
        self.autoScanTv.reset()
        for command in commands:
            check = lkuptable.get(str(command.original_iid), None)
            if check is None:
                logger.debug("ERROR : command without a check")
                continue
            try:
                self.autoScanTv.insert('',"end", str(command.getId()), text=command.name, values=(check.category, check.title, str(check.priority)))
            except tk.TclError as e:
                pass
        




import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.application.scrollabletreeview import ScrollableTreeview
import tkinter as tk

class CheckboxScrollableTreeview(ScrollableTreeview):
    checked_image = None
    unchecked_image = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        bindtags = list(self.treevw.bindtags())
        bindtags.insert(1, self.treevw)
        self.bindtags(tuple(bindtags))
        self.bind("<1>", self.on_sub_click)


    def get_checked_image(self):
        if self.__class__.checked_image is None:
            self.__class__.checked_image = utils.loadIcon("checked.png", resize=(16,16))
        return self.__class__.checked_image

    def get_unchecked_image(self):
        if self.__class__.unchecked_image is None:
            self.__class__.unchecked_image = utils.loadIcon("unchecked.png", resize=(16,16))
        return self.__class__.unchecked_image
    
    def on_sub_click(self, event):
        # Get the clicked coordinates
        x, y = event.x, event.y
        # Identify the item at the clicked coordinates
        item = self.treevw.identify('item', x, y)
        if item is None or item == "":
            return
        column = self.treevw.identify_column(x)
        if column != "#0":
            return
        tags = super().item(item)["tags"]
        tags = [tags] if isinstance(tags, str) else list(tags)
        if 'unchecked' in tags:
            tags.remove('unchecked')
            tags.append('checked')
            image = self.get_checked_image()
        else:
            tags.remove('checked')
            tags.append('unchecked')
            image = self.get_unchecked_image()
        super().item(item, image=image, tags=tags)

    def _check(self, item):
        tags = super().item(item)["tags"]
        tags = [tags] if isinstance(tags, str) else list(tags)
        if 'unchecked' in tags:
            tags.remove('unchecked')
            tags.append('checked')
            image = self.get_checked_image()
            try:
                super().item(item, image=image, tags=tags)
            except tk.TclError:
                pass

    def _uncheck(self, item):
        tags = super().item(item)["tags"]
        tags = [tags] if isinstance(tags, str) else list(tags)
        if 'checked' in tags:
            tags.remove('checked')
            tags.append('unchecked')
            image = self.get_unchecked_image()
            try:
                super().item(item, image=image, tags=tags)
            except tk.TclError:
                pass
    
    def checkAll(self):
        for item in self.get_children(all=True):
            self._check(item)

    def uncheckAll(self):
        for item in self.get_children(all=True):
            self._uncheck(item)        

    def get_checked_children(self):
        return [item for item in self.get_children(all=True) if 'checked' in self.item(item)["tags"]]

    def insert(self, *args, **kwargs):
        kwargs["image"] = self.get_unchecked_image()
        kwargs["tags"] = kwargs.get("tags", []) + ["unchecked"]
        super().insert(*args, **kwargs)


from bson import ObjectId
from pollenisatorgui.core.application.dialogs.ChildDialogAskFile import ChildDialogAskFile
from pollenisatorgui.core.components.apiclient import APIClient
import pollenisatorgui.core.components.utils as utils
import os


def main(apiclient, appli, **kwargs):
    cme_path = utils.which_expand_alias("cme")
    if not cme_path:
        return False, "binary 'cme' is not in the PATH."
    APIClient.setInstance(apiclient)
    ip = ""
    if kwargs.get("target_type").lower() == "share":
        share_info =  apiclient.find("shares", {"type":"share", "_id":ObjectId(kwargs.get("target_iid", ""))}, False)
        if share_info is not None:
            ip = share_info.get("ip", "")
    if ip == "" or ip is None:
        return False, "No ip given"
    files_to_download = share_info.get("infos", {}).get("flagged_files", [])
    export = utils.getExportDir()    
    export_dir = os.path.join(export,share_info.get("ip", ""), share_info.get("share", ""))
    try:
        os.makedirs(export_dir, exist_ok=True)
    except OSError:
        pass
    commands = []
    for file in share_info.get("files" ,[]):
        if file["users"] and file["flagged"]:
            file_name = file["path"].replace(share_info["share"], "", 1)[1:]
            export_path = os.path.join(export_dir, os.path.basename(file_name))
            user_info =  apiclient.find("users", {"type":"user","username":file["users"][0][1], "domain":file["users"][0][0]}, False)
            if user_info and user_info.get("password", "") != "":
                commands.append(f'{cme_path} smb {ip} -d {file["users"][0][0]} -u {file["users"][0][1]} -p {user_info.get("password")} --share {share_info["share"]} --get-file {file_name} {export_path}')
    iid = appli.launch_in_terminal(kwargs.get("default_target",None), "CME download files", " && ".join(commands), use_pollex=False)
    utils.openPathForUser(export_dir)
    
    return True, f"Launched cme in terminal"



"""Defect Model."""

import os
from bson.objectid import ObjectId
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.core.components.apiclient import APIClient


class Defect(Element):
    """
    Represents a Defect object that defines a security defect. A security defect is a note added by a pentester on a port or ip which describes a security defect.

    Attributes:
        coll_name: collection name in pollenisator database
    """
    coll_name = "defects"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched defect is optimal.
                        possible keys with default values are : _id (None), parent (None), tags([]), infos({}),
                        target_id, target_type, title(""), synthesis(""), description(""), ease(""), impact(""), risk(""),
                        redactor("N/A"), type([]), language(""), notes(""), creation_time(None),fixes([]), proofs([]), index(None)
        """
        if valuesFromDb is None:
            valuesFromDb = {}
        if isinstance(valuesFromDb, Defect):
            self = valuesFromDb
            return None
        self.proofs = []
        self.isTemplate = False
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None), valuesFromDb.get("infos", {}))
        types = valuesFromDb.get("type", [])
        if isinstance(types, str):
            types = [types]
        else:
            types = list(types)
        self.initialize(valuesFromDb.get("target_id", ""), valuesFromDb.get("target_type", ""),
                         valuesFromDb.get("title", ""), valuesFromDb.get("synthesis", ""), valuesFromDb.get("description", ""),
                        valuesFromDb.get("ease", ""), valuesFromDb.get(
                            "impact", ""),
                        valuesFromDb.get(
                            "risk", ""), valuesFromDb.get("redactor", "N/A"), types,
                        valuesFromDb.get("language", ""),
                        valuesFromDb.get("notes", ""), valuesFromDb.get("creation_time", None),
                        valuesFromDb.get("fixes", []), valuesFromDb.get("proofs", []), valuesFromDb.get("infos", {}),
                        valuesFromDb.get("index", "0"))

    def initialize(self, target_id="", target_type="", title="", synthesis="", description="", ease="", impact="", risk="", redactor="N/A", mtype=None, language="", notes="", creation_time=None, fixes=None, proofs=None, infos=None, index="0"):
        """Set values of defect
        Args:
            target_id: defect will be assigned to this id; can be empty
            target_type: defect will be assigned to target_id of this type
            title: a title for this defect describing what it is
            synthesis: a short summary of what this defect is about
            description: a more detailed explanation of this particular defect
            ease: ease of exploitation for this defect described as a string 
            impact: impact the defect has on system. Described as a string 
            risk: the combination of impact/ease gives a resulting risk value. Described as a string
            redactor: A pentester that waill be the redactor for this defect.
            mtype: types of this security defects (Application, data, etc...). Default is None
            language: the language in which this defect is redacted
            notes: notes took by pentesters
            proofs: a list of proof files, default to None.
            fixes: a list of fixes for this defect, default to empty list
            infos: a dictionnary with key values as additional information. Default to None
            index: the index of this defect in global defect table (only for unassigned defect)
        Returns:
            this object
        """
        self.title = title
        self.synthesis = synthesis
        self.description = description
        self.ease = ease
        self.impact = impact
        self.risk = risk
        self.redactor = redactor
        self.mtype = mtype if mtype is not None else []
        self.language = language
        self.notes = notes
        self.target_id = target_id
        self.target_type = target_type
        self.infos = infos if infos is not None else {}
        self.proofs = proofs if proofs is not None else []
        self.fixes = fixes if fixes is not None else []
        self.index = index
        self.creation_time = creation_time
        return self

    @classmethod
    def getTemplate(cls, title):
        apiclient = APIClient.getInstance()
        res = apiclient.findInDb("pollenisator", cls.coll_name, {"title":title}, False)
        print("searching "+str(res))
        if res is None:
            return None
        return Defect(res)
    
    @classmethod
    def getTemplateById(cls, iid):
        apiclient = APIClient.getInstance()
        res = apiclient.findDefectTemplateById(ObjectId(iid))
        if res is None:
            return None
        return Defect(res)

    @classmethod
    def getEases(cls):
        """
        Returns: 
            Returns a list of ease of exploitation levels for a security defect.
        """
        return ["Easy", "Moderate", "Difficult", "Arduous", "N/A"]

    @classmethod
    def getImpacts(cls):
        """
        Returns: 
            Returns a list of impact levels for a security defect.
        """
        return ["Critical", "Major", "Important", "Minor", "N/A"]

    @classmethod
    def getRisks(cls):
        """
        Returns: 
            Returns a list of risk levels for a security defect.
        """
        return ["Critical", "Major", "Important", "Minor", "N/A"]

    @classmethod
    def getTypes(cls):
        """
        Returns: 
            Returns a list of type for a security defect.
        """
        return ["Base", "Application", "Policy", "Active Directory", "Infrastructure", "Data"]

    @classmethod
    def getTableRiskFromEase(cls):
        return {"Easy": {"Minor": "Major", "Important": "Major", "Major": "Critical", "Critical": "Critical"},
                             "Moderate": {"Minor": "Important", "Important": "Important", "Major": "Major", "Critical": "Critical"},
                             "Difficult": {"Minor": "Minor", "Important": "Important", "Major": "Major", "Critical": "Major"},
                             "Arduous": {"Minor": "Minor", "Important": "Minor", "Major": "Important", "Critical": "Important"}}
            

    @classmethod
    def getRisk(cls, ease, impact):
        """Dict to find a risk level given an ease and an impact.
        Args:
            ease: ease of exploitation of this defect as as tring
            impact: the defect impact on system security
        Returns:
            A dictionnary of dictionnary. First dict keys are eases of exploitation. Second key are impact strings.
        """
        
        return cls.getTableRiskFromEase().get(ease, {}).get(impact, "N/A")

    def delete(self):
        """
        Delete the defect represented by this model in database.
        """
        ret = self._id
        apiclient = APIClient.getInstance()
        apiclient.delete("defects", ret)

    def addInDefectTemplates(self):
        """
        Add this defect to pollenisator template defect database.
        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        apiclient = APIClient.getInstance()
        base = dict()
        base["title"] = self.title
        base["synthesis"] = self.synthesis
        base["description"] = self.description
        base["ease"] = self.ease
        base["impact"] = self.impact
        base["risk"] = self.risk
        base["type"] = list(self.mtype)
        base["language"] = self.language
        base["fixes"] = self.fixes
        base["perimeter"] = self.perimeter
        res, id = apiclient.insertAsTemplate(base)
        if not res:
            return False, id
        return True, id
        
    def addInDb(self):
        """
        Add this defect to pollenisator database.
        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        apiclient = APIClient.getInstance()
        base = self.getDbKey()
        base["synthesis"] = self.synthesis
        base["description"] = self.description
        base["ease"] = self.ease
        base["impact"] = self.impact
        base["risk"] = self.risk
        base["redactor"] = self.redactor
        base["type"] = list(self.mtype)
        base["language"] = self.language
        base["fixes"] = self.fixes
        base["proofs"] = self.proofs
        base["notes"] = self.notes
        base["fixes"] = self.fixes
        if self.index is not None:
            base["index"] = str(self.index)
        res, id = apiclient.insert("defects", base)
        if not res:
            return False, id
        self._id = id
        return True, id

    def updateAsTemplate(self, pipeline_set=None):
        """Update this object in pollenisator database, template.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        if pipeline_set is None:
            res = apiclient.updateDefectTemplate(ObjectId(self._id), {"title": self.title, "synthesis":self.synthesis, "description":self.description,
                         "ease": self.ease, "impact": self.impact,
                         "risk": self.risk, "type": list(self.mtype), "language":self.language, "fixes":self.fixes})
        else:
            apiclient.updateDefectTemplate(ObjectId(self._id), pipeline_set)

    def update(self, pipeline_set=None):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        if self.isTemplate:
            print("UPDATING TEMPLATE n update"+str(pipeline_set))
            self.updateAsTemplate(pipeline_set)
        if pipeline_set is None:
            apiclient.update("defects", ObjectId(self._id), {"target_id":self.target_id, "target_type":self.target_type, "title": self.title, "synthesis":self.synthesis, "description":self.description,
                         "notes": self.notes, "ease": self.ease, "impact": self.impact,
                         "risk": self.risk, "redactor": self.redactor, "type": list(self.mtype), "language":self.language, "proofs": self.proofs, "fixes":self.fixes, "infos": self.infos, "index":str(self.index)})
        else:
            apiclient.update("defects", ObjectId(self._id), pipeline_set)

    def _getParentId(self):
        """
        Return the mongo ObjectId _id of the first parent of this object. For a Defect it is either an ip or a port depending on the Defect's level.

        Returns:
            Returns the parent's ObjectId _id".
        """
        return self.target_id

    def calcDirPath(self):
        """Returns a directory path constructed for this defect.
        Returns:
            path as string
        """
        apiclient = APIClient.getInstance()
        path_calc = str(apiclient.getCurrentPentest())+"/"+str(self.target_id)
        return path_calc

    def uploadProof(self, proof_local_path):
        """Upload the given proof file to the server
        Args:
            proof_local_path: a path to a local proof file
        Returns:
            the basename of the file 
        """
        apiclient = APIClient.getInstance()
        result = apiclient.putProof(self._id, proof_local_path)
        if result.get("remote_path", "") != "":
            return result.get("remote_path", "")
        return None

    def getProof(self, ind):
        """Download the proof file at given proof index
        Returns:
            A string giving the local path of the downloaded proof
        """
        apiclient = APIClient.getInstance()
        current_dir = os.path.dirname(os.path.realpath(__file__))
        local_path = os.path.join(current_dir, "../../results", self.calcDirPath())
        try:
            os.makedirs(local_path)
        except FileExistsError:
            pass
        local_path = os.path.join(local_path, self.proofs[ind])
        ret = apiclient.getProof(self._id, self.proofs[ind], local_path)
        return ret
    
    def getProofWithName(self, name):
        apiclient = APIClient.getInstance()
        current_dir = os.path.dirname(os.path.realpath(__file__))
        local_path = os.path.join(current_dir, "../../results", self.calcDirPath())
        try:
            os.makedirs(local_path)
        except FileExistsError:
            pass
        ret = apiclient.getProof(self.getId(), name, local_path)
        return ret

    def removeProof(self, ind):
        """Removes the proof file at given proof index
        """
        apiclient = APIClient.getInstance()
        filename = self.proofs[ind]
        ret = apiclient.rmProof(self._id, filename)
        del self.proofs[ind]
        return ret

    def __str__(self):
        """
        Get a string representation of a defect.

        Returns:
            Returns the defect +title.
        """
        return self.title
    
    def getTargetRepr(self):
        apiclient = APIClient.getInstance()
        result = apiclient.getDefectTargetRepr([self.getId()])
        if result is not None:
            return result.get(str(self.getId()), None)
        return None

    def getDetailedString(self, onlyTarget=False):
        """Returns a detailed string describing for this defect.
        Returns:
            the defect title. If assigned, it will be prepended with ip and (udp/)port
        """
        ret = ""
        target = self.getTargetRepr()
        if onlyTarget:
            return target
        ret = target+" "+self.__str__()
        return ret

    def getDbKey(self):
        """Return a dict from model to use as unique composed key.
        Returns:
            A dict (3 keys :"target_id" and "target_type" and "title")
        """
        return {"target_type": self.target_type, "target_id": self.target_id, "title": self.title}

    def isAssigned(self):
        """Returns a boolean indicating if this defect is assigned to an ip or is global.
        Returns:
            bool
        """
        return self.target_id != ""

    @classmethod
    def getDefectTable(cls):
        """Return the table of global defects sorted by their index field
        Returns:
            A list of Defect
        """
        return APIClient.getInstance().getDefectTable()

    @classmethod
    def fetchPentestObjects(cls):
        apiclient = APIClient.getInstance()
        ds = apiclient.find(cls.coll_name, {"target_id":{"$ne":""}}, True)
        if ds is None:
            return None
        for d in ds:
            # disabling this error as it is an abstract function
            yield cls(d)  

    def getData(self):
        """Return defect attributes as a dictionnary matching Mongo stored defects
        Returns:
            dict with keys title, ease, ipact, risk, redactor, type, notes, ip, port, proto, proofs, _id, tags, infos
        """
        return {"title": self.title, "synthesis":self.synthesis, "description":self.description, "ease": self.ease, "impact": self.impact,
                "risk": self.risk, "redactor": self.redactor, "type": self.mtype, "language":self.language, "notes": self.notes, "fixes":self.fixes,
                "target_id": self.target_id, "target_type": self.target_type,"index":self.index, "creation_time":self.creation_time,
                "proofs": self.proofs, "_id": self.getId(),"infos": self.infos}



"""Ask a question to the user.
"""
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import pollenisatorgui.core.components.utils as utils

class ChildDialogQuestion:
    """
    Open a child dialog of a tkinter application to ask a question.
    """

    def __init__(self, parent, title, question, answers=("Yes", "No")):
        """
        Open a child dialog of a tkinter application to ask a question.

        Args:
            parent: the tkinter parent view to use for this window construction.
            title: title of the new window
            question: question to answer
            answers: a tuple with possible answers. Default to ("Yes" ,"No")
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.title(title)
        self.app.resizable(True, True)
        
        appFrame = CTkFrame(self.app)
        self.rvalue = None
        self.parent = parent
        lbl = CTkLabel(appFrame, text=question)
        lbl.pack(side=tk.TOP, padx=10, pady=10, fill=tk.X)
        buttonsFrame = CTkFrame(appFrame)
        for i, answer in enumerate(answers):
            if answer == "Cancel" or answer == "No":
                self.cancel_button = CTkButton(buttonsFrame, text=answer, fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                                border_width=1, border_color="firebrick1", hover_color="tomato")
                self.app.bind("<Escape>", self.escape)
                self.cancel_button.bind("<Button-1>", self.onOk)
                self.cancel_button.grid(row=0, column=i, padx=15)
            else:
                _button = CTkButton(buttonsFrame, text=answer)
                _button.bind("<Button-1>", self.onOk)
                _button.grid(row=0, column=i, padx=15)

        buttonsFrame.pack(side=tk.TOP, ipadx=5, pady=5)
        appFrame.pack(fill=tk.BOTH)
        _button.focus_set()
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.focus_force()
            self.app.grab_set()
            self.app.lift()
        except tk.TclError:
            pass

    def escape(self, _event=None):
        self.rvalue = self.cancel_button.cget("text")
        self.quit()
        
    def onOk(self, event):
        """
        Called when the user clicked the validation button.
        Set the rvalue attributes to the answer string choosen.
        """
        # send the data to the parent
        widget = event.widget.master
        self.rvalue = widget.cget("text")
        self.quit()

    def quit(self):
        self.app.destroy()
